from webform.models.definition import Action
import base64


class Login:
    def __init__(self, rcic_account):
        self.rcic_account = rcic_account

    def encode(self, password):
        password_bytes = password.encode("ascii")
        base64_bytes = base64.b64encode(password_bytes)
        return base64_bytes.decode("ascii")

    def login(self):
        return [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Sign in",
                "actions": [
                    {
                        "action_type": Action.GotoPage.value,
                        "url": "https://prson-srpel.apps.cic.gc.ca/en/rep/login",
                    },
                    {
                        "action_type": Action.Login.value,
                        "label": "Login",
                        "account": self.rcic_account["account"],
                        "password": self.encode(self.rcic_account["password"]),
                        "account_element_id": "#username",
                        "password_element_id": "#password",
                        "login_button_id": "body > pra-root > pra-localized-app > main > div > pra-login-page > pra-login > div > div > form > button",
                    },
                ],
                "id": None,
            }
        ]



from datetime import datetime
import time

from .form_controls import (
    Control,
    TextField,
    RadioButton,
    CheckBox,
    DropdownList,
    DateField,
    Skip,
    Pause,
    OutputInfo,
    Button,
)


class ApplicationForm(object):
    """Application PDF form"""

    def __init__(self, applicant, operations=None):
        self.applicant = applicant
        if operations is not None:
            self._operations = list(operations)
        else:
            self._operations = []

    def add_step(self, operation: Control):
        self._operations.append(operation)

    def add_text(self, data, pause=0):
        control = TextField(data, pause=pause)
        self.add_step(control)

    def add_radio(self, data):
        control = RadioButton(data)
        self.add_step(control)

    def add_checkbox(self, data):
        control = CheckBox(data)
        self.add_step(control)

    def add_dropdown(self, data, options, like: bool = False):
        control = DropdownList(data, options, like)
        self.add_step(control)

    def add_date(self, date, noday: bool = False, pause=0.2):
        self.add_step(DateField(date, noday, pause=pause))

    def add_skip(self, num):
        """Add {num} skips"""
        for _ in range(num):
            self.add_step(Skip())

    def add_pause(self, seconds):
        """Pause {seconds} seconds"""
        self.add_step(Pause(seconds))

    def add_button(self):
        """Click button"""
        self.add_step(Button())

    def add_info(self, info: str):
        """Output info for debugging script"""
        self.add_step(OutputInfo(info))

    def fill_form(self, start: int = 0, verbose: bool = False):
        """Perform the fill pdf form action"""
        step = 0
        for operation in self._operations[start:]:
            # if not isinstance(operation, OutputInfo):
            if verbose:
                operation.verbose = True
                print(step, end="\t")
                step += 1
            operation.fill()



from termcolor import colored
from utils.utils import checkContinuity
from typing import List, Optional, Union
from model.common.address import Address
from model.common.educationbase import EducationBase
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from model.common.id import ID
from model.common.person import Person
from datetime import date
from pydantic import BaseModel, EmailStr, root_validator, validator
from model.common.utils import normalize
from webform.prportal.data.country_residence import country_residence


class PersonId(ID):
    pass


class PRCase(BaseModel):
    imm_program: str
    imm_category: str
    imm_under: Optional[str]
    communication_language: str
    interview_language: str
    need_translator: bool
    intended_province: str
    intended_city: str


class COR(BaseModel):
    start_date: date
    end_date: date
    country: str
    status: str
    explanation: Optional[str]

    @validator("country")
    def must_in_list(cls, v):
        if v not in country_residence.keys():
            similars = []
            # 全部变小写比较
            for key in country_residence.keys():
                if key.lower() == v.lower():
                    raise ValueError(
                        f"Please use correct upper letter. The country is {key}, but your input is {v}"
                    )
                elif v.lower() in key.lower():
                    similars.append(f"Possible matching country {key}")
            if similars:
                raise ValueError(
                    f"Your input {v} may matches:",
                    similars,
                    "Please correct it in your excel table-cor",
                )
        return v


class CORs(object):
    def __init__(self, cors: List[COR]):
        self.cors = cors

    @property
    def current(self):
        if len(self.cors) >= 1:
            return self.cors[0]
        else:
            raise ValueError(
                f"table-cor must have values, and the first line must be current residence country."
            )


class Marriage(BaseModel):
    marital_status: str
    married_date: Optional[date]
    sp_last_name: Optional[str]
    sp_first_name: Optional[str]
    previous_married: bool
    pre_sp_last_name: Optional[str]
    pre_sp_first_name: Optional[str]
    pre_sp_dob: Optional[date]
    pre_relationship_type: Optional[str]
    pre_start_date: Optional[date]
    pre_end_date: Optional[date]


class Personal(Person):
    used_last_name: Optional[str]
    used_first_name: Optional[str]
    native_last_name: str
    native_first_name: str
    height: int
    eye_color: str
    country_of_birth: str
    place_of_birth: str
    uci: Optional[str]
    citizen2: Optional[str]
    native_language: str
    english_french: str
    which_one_better: Optional[str]
    language_test: bool
    current_occupation: str
    intended_occupation: Optional[str]
    email: EmailStr
    primary_school_years: int
    secondary_school_years: int
    post_secondary_school_years: int
    other_school_years: int
    last_entry_date: Optional[date]
    last_entry_place: Optional[str]
    other_explanation: Optional[str]
    accompany_to_canada: bool
    relationship_to_pa: Optional[str]
    dependant_type: Optional[str]

    _normalize_used_first_name = validator(
        "used_first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_used_last_name = validator(
        "used_last_name", allow_reuse=True, check_fields=False
    )(normalize)

    @root_validator
    def checkAnswers(cls, values):
        questions = ["english_french"]
        explanations = ["which_one_better"]
        qas = dict(zip(questions, explanations))
        for k, v in qas.items():
            if values.get(k) and values.get(k) == "Both" and not values.get(v):
                raise ValueError(
                    f"Since {k} is true, but you did not answer the question {v} in info-personal sheet"
                )
        return values


class Education(EducationBase):
    city: Optional[str]
    country: Optional[str]


class Family(BaseModel):
    relationship: Optional[str]
    last_name: Optional[str]
    first_name: Optional[str]
    native_last_name: Optional[str]
    native_first_name: Optional[str]
    date_of_birth: Optional[date]
    date_of_death: Optional[date]
    place_of_birth: Optional[str]
    birth_country: Optional[str]
    marital_status: Optional[str]
    email: Optional[EmailStr]
    address: Optional[str]


class Travel(BaseModel):
    start_date: date
    end_date: date
    length: int
    destination: str
    purpose: str


class PRBackground(BaseModel):
    q1: bool
    q2: bool
    q3: bool
    q4: bool
    q5: bool
    q6: bool
    q7: bool
    q8: bool
    q9: bool
    q10: bool
    q11: bool
    details: Optional[str]


class History(BaseModel):
    start_date: date
    end_date: Optional[date]
    activity: str
    city_and_country: str
    status: str
    name_of_company_or_school: str
    
    @property
    def __str__(self):
        return "table-history"


class Member(BaseModel):
    start_date: date
    end_date: date
    organization_name: str
    organization_type: str
    position: str
    city: str
    country: str


class Government(BaseModel):
    start_date: date
    end_date: date
    country: str
    department: str
    position: str


class Military(BaseModel):
    start_date: date
    end_date: date
    country: str
    service_detail: str
    rank: str
    combat_detail: Optional[str]
    reason_for_end: Optional[str]


class AddressHistory(BaseModel):
    start_date: date
    end_date: Optional[date]
    street_and_number: str
    city: str
    province: str
    country: str
    post_code: str
    
    @property
    def __str__(self):
        return "table-addresshistory"


class PrModel(CommonModel):
    personal: Personal
    cor: List[COR]
    marriage: Marriage
    address: List[Address]
    phone: List[Phone]
    personid: List[PersonId]
    prcase: PRCase
    family: List[Family]
    travel: List[Travel]
    prbackground: PRBackground
    education: List[Education]
    history: List[History]
    member: List[Member]
    government: List[Government]
    military: List[Military]
    addresshistory: List[AddressHistory]

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None,check=True):
        if output_excel_file:
            excels=self.getExcels(['excel/pr.xlsx','excel/pa.xlsx'])
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())
        if check: [self.check(item,excels[0]) for item in [self.history,self.addresshistory]]
    
    def check(self,items:List[object],excel_file):
        results=[]
        # construct list suitable for checkContinuity
        for item in items:
            results.append(list(item.__dict__.values()))
        
        # check 
        continued, sorted_list,msg=checkContinuity(results)
        
        if not continued:
            print(colored(f"There are {len(msg)} error(s) in sheet {items[0].__str__}","red"))
            [print(index,"\t",m) for index,m in enumerate(msg)]



from datetime import date,datetime
from pydantic import BaseModel
from typing import Union,Optional

class Language(BaseModel):
    reading:Union[int,float]
    writting:Union[int,float]
    listening:Union[int,float]
    speaking:Union[int,float]
    test_date:date
    sign_data:date

    def is_valid(self,the_day=date.today()):
        if not isinstance(the_day,date):
            the_day=datetime.strptime(the_day,"%Y-%m-%d").date()
        return True if (the_day-self.sign_data).days<730 else False
    def __str__(self):
        return f"{self.reading},{self.writting},{self.listening},{self.speaking}"
    class Meta:
        abstract=True
        
class CLB():
    # 1. IELTS->CLB level, Reading	Writing	Listening	Speaking
    ielts_table = [
        [10, 8.0, 7.5, 8.5, 7.5],
        [9, 7.0, 7.0, 8.0, 7.0],
        [8, 6.5, 6.5, 7.5, 6.5],
        [7, 6.0, 6.0, 6.0, 6.0],
        [6, 5.0, 5.5, 5.5, 5.5],
        [5, 4.0, 5.0, 5.0, 5.0],
        [4, 3.5, 4.0, 4.5, 4.0]
    ]
    tef_table = [
        [10, 263, 277, 393, 415, 316, 333, 393, 415],
        [9, 248, 262, 371, 392, 298, 315, 371, 392],
        [8, 233, 247, 349, 370, 280, 297, 349, 370],
        [7, 207, 232, 310, 348, 249, 279, 310, 348],
        [6, 181, 206, 271, 309, 217, 248, 271, 309],
        [5, 151, 180, 226, 270, 181, 216, 226, 270],
        [4, 121, 150, 181, 225, 145, 180, 181, 225]
    ]
    tcf_table = [
        [10, 549, 699, 16, 20, 549, 699, 16, 20],
        [9, 524, 548, 14, 15, 523, 548, 14, 15],
        [8, 499, 523, 12, 13, 503, 522, 12, 13],
        [7, 453, 498, 10, 11, 458, 502, 10, 11],
        [6, 406, 452, 7, 9, 398, 457, 7, 9],
        [5, 375, 405, 6, 6, 369, 397, 6, 6],
        [4, 342, 374, 4, 5, 331, 368, 4, 5]
    ]

    def __init__(self, clb_level):
        self.level=clb_level
        self.level=10 if self.level>10 else self.level

    # input CLB level, return r w l s of IELTS
    def to_ielts(self):
        if self.level<4:
            return [0,0,0,0]
        index=10-self.level
        return CLB.ielts_table[index][1:5]
    
    # 输入CLB值 返回对应的tef r w l s分数列表
    def to_tef(self):
        if self.level<4:
            return [0,0,0,0,0,0,0,0]
        index=10-self.level
        return CLB.tef_table[index][1:9]
    
        # 输入CLB值 返回对应的tef r w l s分数列表
    def to_tcf(self):
        if self.level<4:
            return [0,0,0,0,0,0,0,0]
        index=10-self.level
        return CLB.tcf_table[index][1:9]

    
class IELTS(Language):
    TYPE=[
        ('General','General'),
        ('Academic','Academic')
    ]
    reading:float
    writting:float
    listening:float
    speaking:float
    type_of_test:str
    report_number:str

    def __init__(self,reading=None,writting=None,listening=None,speaking=None):
        self.reading=reading
        self.writting=writting
        self.listening=listening
        self.speaking=speaking


    def _ielts_2_clbs(self):
        ielts=[self.reading,self.writting,self.listening,self.speaking]
        clb_level = []
        for column in range(1, 5):  # 从读到说，4个列，从1开始，开始找对应的clb级别
            factor = float(ielts[column - 1])  # r w l s每个子项
            for level in range(0, 7):  # 从CLB 级别10一直找到4 7行
                if factor > CLB.ielts_table[0][column]:  # 如果语言成绩高于最高分，按照最高分算
                    clb_level.append(10) 
                    break
                if factor < CLB.ielts_table[6][column]:  # 如果低于最低分，则CLB Level =0 跳出本列循环
                    clb_level.append(0)
                    break
                # 要算每个子项的CLB值
                if CLB.ielts_table[level][column] == ielts[column - 1]:
                    clb_level.append(CLB.ielts_table[level][0])  # 得到该雅思成绩对应的CLB 级别
                    break
            # 如果某项分数不在阵列中
            for level in range(1, 7):
                if CLB.ielts_table[level][column] < factor < CLB.ielts_table[level - 1][column]:
                    clb_level.append(CLB.ielts_table[level][0])  # 得到该雅思成绩对应的CLB 级别
                    break
        return clb_level
    
    # public class mothod for returning single factor's clb level. reading:0, writing:1, listening:2, speaking:3
    @classmethod
    def ielts_2_clb(self,score,factor_index):
        column=factor_index+1
        for level in range(0, 7):  # 从CLB 级别10一直找到4 7行
                if score > CLB.ielts_table[0][column]:  # 如果语言成绩高于最高分，按照最高分算
                    return 10
                if score < CLB.ielts_table[6][column]:  # 如果低于最低分，则CLB Level =0 跳出本列循环
                    return 0
                # 要算每个子项的CLB值
                if CLB.ielts_table[level][column] == score:
                    return CLB.ielts_table[level][0]  # 得到该雅思成绩对应的CLB 级别
        # 如果某项分数不在阵列中
        for level in range(1, 7):
            if CLB.ielts_table[level][column] < score < CLB.ielts_table[level - 1][column]:
                return CLB.ielts_table[level][0]  # 得到该雅思成绩对应的CLB 级别
    

    @property
    def clb_r(self):
        return self._ielts_2_clbs()[0]
    
    @property
    def clb_w(self):
        return self._ielts_2_clbs()[1]
    
    @property
    def clb_l(self):
        return self._ielts_2_clbs()[2]
    
    @property
    def clb_s(self):
        return self._ielts_2_clbs()[3]

    @property
    def clb(self):
        return min(self.clb_r,self.clb_w,self.clb_l,self.clb_s)

class CELPIP(Language):
    TYPE=[
        ('General','General'),
        ('Academic','Academic')
    ]
    registration_number:str
    pin_number:str

    def __init__(self,reading=None,writting=None,listening=None,speaking=None):
        self.reading=reading
        self.writting=writting
        self.listening=listening
        self.speaking=speaking

    @property
    def clb_r(self):
        return self.reading
    
    @property
    def clb_w(self):
        return self.writting
    
    @property
    def clb_l(self):
        return self.listening
    
    @property
    def clb_s(self):
        return self.speaking

    @property
    def clb(self):
        return min(self.reading,self.writting,self.listening,self.speaking)

class TEF(Language):
    def __init__(self,reading=None,writting=None,listening=None,speaking=None):
        self.reading=reading
        self.writting=writting
        self.listening=listening
        self.speaking=speaking

    def _tef_2_clbs(self):
        if self.reading not in range(121,278):
            print('TEF reading score is invalid')
            exit(1)
        if self.writting not in range(181,416):
            print('TEF writing score is invalid')
            exit(1)
        if self.listening not in range(145,334):
            print('TEF listening score is invalid')
            exit(1)
        if self.speaking not in range(181,416):
            print('TEF speaking score is invalid')
            exit(1)
        rwls=[self.reading,self.writting,self.listening,self.speaking]
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(CLB.tef_table[row][start], CLB.tef_table[row][end] + 1):
                    clb.append(CLB.tef_table[row][0])
                    break
        return clb

    @property
    def clb_r(self):
        return self._tef_2_clbs()[0]
    
    @property
    def clb_w(self):
        return self._tef_2_clbs()[1]
    
    @property
    def clb_l(self):
        return self._tef_2_clbs()[2]
    
    @property
    def clb_s(self):
        return self._tef_2_clbs()[3]

    @property
    def clb(self):
        return min(self.clb_r,self.clb_w,self.clb_l,self.clb_s)


class TCF(Language):
    def __init__(self,reading=None,writting=None,listening=None,speaking=None):
        self.reading=reading
        self.writting=writting
        self.listening=listening
        self.speaking=speaking

    def _tcf_2_clbs(self):
        if self.reading not in range(342,700):
            print('TCF reading score is invalid')
            exit(1)
        if self.writting not in range(4,21):
            print('TCF writing score is invalid')
            exit(1)
        if self.listening not in range(331,700):
            print('TCF listening score is invalid')
            exit(1)
        if self.speaking not in range(4,21):
            print('TCF speaking score is invalid')
            exit(1)
        rwls=[self.reading,self.writting,self.listening,self.speaking]
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(CLB.tcf_table[row][start], CLB.tcf_table[row][end] + 1):
                    clb.append(CLB.tcf_table[row][0])
                    break
        return clb
    
    @property
    def clb_r(self):
        return self._tcf_2_clbs()[0]
    
    @property
    def clb_w(self):
        return self._tcf_2_clbs()[1]
    
    @property
    def clb_l(self):
        return self._tcf_2_clbs()[2]
    
    @property
    def clb_s(self):
        return self._tcf_2_clbs()[3]

    @property
    def clb(self):
        return min(self.clb_r,self.clb_w,self.clb_l,self.clb_s)


from email.errors import ObsoleteHeaderDefect
from optparse import Option
from typing import List, Optional, Union
from model.common.address import Address
from model.common.educationbase import EducationBase
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from model.common.employmentbase import EmploymentBase
from model.common.jobofferbase import JobofferBase
from model.common.id import ID
from model.common.person import Person
from model.common.cor import CORs, COR
from model.common.status import Status
from datetime import date, datetime
from pydantic import BaseModel, EmailStr, root_validator, validator
from model.common.utils import normalize
from webform.prportal.data.country_residence import country_residence
from model.common.contact import ContactBase
from model.common.family import FamilyBase


class PersonId(ID):
    pass


class CanadaRelative(BaseModel):
    last_name: Optional[str]
    first_name: Optional[str]
    sex: Optional[str]
    relationship: Optional[str]
    status: Optional[str]
    age: Optional[int]
    city: Optional[str]
    province: Optional[str]
    years_in_canada: Optional[int]


class Rcic(BaseModel):
    first_name: Optional[str]
    last_name: Optional[str]
    telephone: Optional[str]


class General(BaseModel):
    legal_name: str
    operating_name: Optional[str]
    corporate_structure: str
    other_explaination: Optional[str]
    establish_date: date
    industry: str
    registration_number: str
    ft_employee_number: int
    establish_date: date
    website: Optional[str]


class Contact(ContactBase):
    position: str


class ErAddress(Address):
    phone: Optional[str]


class Joboffer(JobofferBase):
    phone: str
    is_working: bool
    license_request: bool


class Employment(EmploymentBase):
    bcpnp_qualified: bool
    phone_of_certificate_provider: Optional[str]
    website: Optional[str]
    unit: Optional[str]
    street_address: Optional[str]
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]
    postcode: Optional[str]
    duties: Optional[str]


class Marriage(BaseModel):
    marital_status: Optional[str]
    married_date: Optional[date]
    sp_last_name: Optional[str]
    sp_first_name: Optional[str]
    sp_in_canada: Optional[bool]
    sp_canada_status: Optional[str]
    sp_canada_status_end_date: Optional[date]
    sp_in_canada_other: Optional[str]
    sp_in_canada_work: Optional[bool]
    sp_canada_occupation: Optional[str]
    sp_canada_employer: Optional[str]


class Personal(Person):
    uci: Optional[str]

    _normalize_used_first_name = validator(
        "used_first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_used_last_name = validator(
        "used_last_name", allow_reuse=True, check_fields=False
    )(normalize)

    @property
    def user_id(self):
        dob = self.dob.strftime(("%Y-%m-%d"))
        return (
            self.last_name[0].upper()
            + self.first_name[0]
            + dob.split("-")[0]
            + dob.split("-")[1]
            + dob.split("-")[2]
        )

    @property
    def password(self):
        return "Super" + str(datetime.today().year) + "!"


class Education(EducationBase):
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]
    is_trade: Optional[bool]
    graduate_date: Optional[date]


class Family(FamilyBase):
    country_of_citizenship: Optional[str]
    address: Optional[str]
    date_of_death: Optional[date]
    marital_status: Optional[str]


class Bcpnp(BaseModel):
    intended_city: str
    account: Optional[str]
    password: Optional[str]
    # has_current_app: bool
    # has_applied_before: bool
    # pre_file_no: Optional[str]
    case_stream: Optional[str]
    q1: bool
    q1_explaination: Optional[str]
    q2: bool
    q2_explaination: Optional[str]
    q3: bool
    q3_explaination: Optional[str]
    q4: bool
    q4_file_number: Optional[str]
    q4_explaination: Optional[str]
    q5: bool
    q5_explaination: Optional[str]
    q6: bool
    q6_explaination: Optional[str]
    q7: bool
    q7_explaination: Optional[str]


class Ee(BaseModel):
    ee_profile_no: str
    ee_expiry_date: date
    ee_jsvc: str
    ee_score: str
    ee_noc: str
    ee_job_title: str


class BcpnpModelApp(CommonModel):
    personal: Personal
    marriage: Marriage
    cor: List[COR]
    status: Status
    general: General
    joboffer: Joboffer
    family: List[Family]
    eraddress: List[ErAddress]
    contact: List[Contact]
    bcpnp: Bcpnp
    education: List[Education]
    employment: List[Employment]
    rcic: Rcic
    canadarelative: List[CanadaRelative]

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(
                ["excel/er.xlsx", "excel/pa.xlsx", "excel/bcpnp.xlsx", "excel/rep.xlsx"]
            )
        else:
            if excels is None or len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())


class BcpnpEEModelApp(BcpnpModelApp):
    ee: Ee

    def __init__(self, excels=None, output_excel_file=None):
        super().__init__(excels=excels, output_excel_file=output_excel_file)



from pydantic import validator,EmailStr
from typing import Optional,List
from model.common.employerbase import EmployerBase
from model.common.jobofferbase import JobofferBase
from model.common.address import Address,Addresses
from model.common.utils import makeList


class General(EmployerBase):
    company_intro:str
    business_intro:str
    recruit_email:EmailStr
        
class JobOffer(JobofferBase):
    # education_level:Optional[str]
    disability_insurance:bool
    dental_insurance:bool
    empolyer_provided_persion:bool
    extended_medical_insurance:bool
    extra_benefits:Optional[str]
    duties_brief:str
    duties:list
    specific_edu_requirement:str
    skill_experience_requirement:str
    other_requirements:Optional[list]

    _str2bool_duties=validator('duties',allow_reuse=True,pre=True)(makeList)
    _str2bool_other_requirements=validator('other_requirements',allow_reuse=True,pre=True)(makeList)
    
    @property
    def benefits(self):
        bs=[]
        if self.disability_insurance:bs.append('Disability insurance')
        if self.dental_insurance:bs.append('Dental insurance')
        if self.empolyer_provided_persion:bs.append('Employer provided persion')
        if self.extended_medical_insurance:bs.append('Extended medical insurance')
        if self.extra_benefits: bs.append(self.extra_benefits)
        return ', '.join(bs)

class ErAddress(Address):
    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()
    
class ErAddresses(Addresses):
    def __init__(self, address_list: List[Address]) -> None:
        super().__init__(address_list)

    







from pdfform import context
from model.tr.m1295 import M1295Model
from model.tr.m5708 import M5708Model
from model.tr.m5709 import M5709Model
from model.tr.m5710 import M5710Model

test_1295 = {
    ## ------- header -------
    "uci": "91075022",
    "service_language": "English",
    ## ------ personal details ------
    "last_name": "ZHANG",
    "first_name": "SAN",
    #'used_first_name': '',
    #'used_last_name': '',
    "sex": "Female",
    "dob": "1976-06-29",
    "city_of_birth": "SHANGHAI",
    "country_of_birth": "China",
    "citizenship": "China",
    "current_residence": {
        "country": "China",
        "status": "Citizen",
        "other": "",
        "from": "",
        "to": "",
    },
    #'previous_residence': [],
    #'applying_country': {},
    "current_mariage": {
        "status": "Married",
        "married_date": "2004-02-14",
        "spouse_last_name": "LI",
        "spouse_first_name": "SI",
    },
    "previous_mariage": {
        "spouse_last_name": "WANG",
        "spouse_first_name": "WU",
        "spouse_dob": "1972-07-04",
        "type": "Married",
        "from": "1999-10-28",
        "to": "2000-10-20",
    },
    ## ------ language ------
    "native_language": "Mandarin",
    "english_french": "Neither",
    #'preferred_language': '',
    "language_test": "No",
    ## ------ passport ------
    "passport": {
        "number": "EJ4484570",
        "country_of_issue": "China",
        "issue_date": "2021-02-25",
        "expiry_date": "2031-02-24",
        #'taiwan_issued': '',
        #'israel_issued': '',
    },
    ## ------ national ID ------
    "national_id": {
        "number": "42010419760629202X",
        "country_of_issue": "CHN (China)",  # TODO: how to convert to normal country name?
        "issue_date": "2005-12-20",
        "expiry_date": "2025-12-20",
    },
    ## ------ US PR card ------
    # "us_pr_card": {
    #     "number": "",
    #     "expiry_date": "",
    # },
    ## ------ contact information ------
    "mailing_address": {
        "po_box": "",
        "apt": "41-501",
        "street_no": "380",
        "street_name": "XINGZHI ROAD",
        "city": "SHANGHAI",
        "country": "China",
        "province": "",
        "postal_code": "200442",
        "district": "BAOSHAN",
    },
    #'residential_address': {},
    "telephone_no": {
        "canada": "No",
        "type": "Cellular",
        "country_code": "86",
        "number": "18917872110",
        "ext": "",
    },
    #'alternate_no': {},
    #'fax_no': {},
    "email": "yaozhenyu2005@gmail.com",
    ## ------ work detail ------
    "work_detail": {
        "type": "Open Work Permit",
        #'employer_name': 'Not applicable',
        #'employer_address': 'Not applicable',
        "province": "BC",
        "city": "Nanaimo",
        #'address': 'TO BE DETERMINED',
        #'job_title': 'TO BE DETERMINED',
        #'description': 'TO BE DETERMINED',
        "from": "2022-07-01",
        "to": "2024-05-01",
        #'lmia_no': '',
    },
    ## ------ education ------
    "education": {
        "from": "2002-09",
        "to": "2005-01",
        "field": "Accounting, Diploma",
        "school": "China Central Radio & TV University",
        "city": "WUHAN",
        "country": "China",
        #'province': '',
    },
    ## ------ Employment ------
    "employment": [
        {
            "from": "2012-02",
            "to": "2021-11",
            "occupation": "PRESIDENTT, OWNER-GENERAL-MANAGER",
            "company": "SHANGHAI CHONGYUAN MEDICAL DEVICE CO.,LTD",
            "city": "SHANGHAI",
            "country": "China",
            #'province': '',
        },
        {
            "from": "2006-05",
            "to": "2012-01",
            "occupation": "OWNER-GENERAL-MANAGER",
            "company": "SHANGHAI CHONGYUAN MEDICAL DEVICE CO.,LTD",
            "city": "SHANGHAI",
            "country": "China",
            #'province': '',
        },
    ],
}

test_5708 = {
    ## ------- header -------
    "uci": "1120956934",
    "service_language": "English",
    "extend_visitor": True,
    "restore_visitor": False,
    "new_permit": False,
    ## ------ personal details ------
    "last_name": "JIA",
    "first_name": "LIU",
    "used_first_name": "JIA",
    "used_last_name": "QI",
    "sex": "Male",
    "dob": "1971-10-28",
    "city_of_birth": "SHANGHAI",
    "country_of_birth": "China",
    "citizenship": "China",
    "current_residence": {
        "country": "Canada",
        "status": "Visitor",
        "other": "",
        "from": "2020-08-17",
        "to": "2022-08-06",
    },
    #'previous_residence': [],
    #'applying_country': {},
    "current_mariage": {
        "status": "Married",
        "married_date": "2004-02-14",
        "spouse_last_name": "LI",
        "spouse_first_name": "YIN",
        "spouse_is_canadian": False,
    },
    # "previous_mariage": {},
    ## ------ language ------
    "native_language": "Mandarin",
    "english_french": "English",
    #'preferred_language': '',
    "language_test": "No",
    ## ------ passport ------
    "passport": {
        "number": "EJ4692323",
        "country_of_issue": "China",
        "issue_date": "2021-04-01",
        "expiry_date": "2031-03-31",
        #'taiwan_issued': '',
        #'israel_issued': '',
    },
    ## ------ national ID ------
    "national_id": {
        "number": "310108197110285213",
        "country_of_issue": "CHN (China)",  # TODO: how to convert to normal country name?
        "issue_date": "",
        "expiry_date": "",
    },
    ## ------ US PR card ------
    # "us_pr_card": {
    #     "number": "",
    #     "expiry_date": "",
    # },
    ## ------ contact information ------
    "mailing_address": {
        "po_box": "",
        "apt": "",
        "street_no": "6499",
        "street_name": "RAVEN RD",
        "city": "NANAIMO",
        "country": "Canada",
        "province": "BC",
        "postal_code": "V9V 1V7",
    },
    #'residential_address': {},
    "telephone_no": {
        "canada": "No",
        "type": "Cellular",
        "country_code": "86",
        "number": "18917872110",
        "ext": "",
    },
    #'alternate_no': {},
    #'fax_no': {},
    "email": "",
    ## ------ coming into Canada ------
    "coming_into_canada": {
        "date": "2020-01-02",
        "place": "Vancouver",
        "purpose": "Family Visit",
        #'other': '',
        #'last_date': '',
        #'last_place': '',
        #'last_document_number': '',
    },
    ## ------ details of visit to Canada ------
    "visit_detail": {
        "purpose": "Family Visit",
        #'other': '',
        "from": "2022-06-15",
        "to": "2023-09-30",
        "funds": "15000",
        "expense_by": "Myself",
        #'other': '',
        "persons": [
            {
                "name": "YAO ZICHENG",
                "relation": "",
                "address": "6499 RAVEN RD NANAIMO BC  V9V 1V7",
            },
        ],
    },
    ## ------ education ------
    "education": {
        "from": "1997-04",
        "to": "2004-12",
        "field": "DIPLOMA",
        "school": "FUDAN UNIVERSITY",
        "city": "SHANGHAI",
        "country": "China",
        #'province': '',
    },
    ## ------ Employment ------
    "employment": [
        {
            "from": "2012-02",
            "to": "2020-08",
            "occupation": "MANAGER",
            "company": "SHANGHAI CHONGYUAN MEDICAL DEVICE CO.,LTD",
            "city": "SHANGHAI",
            "country": "China",
            #'province': '',
        },
        {
            "from": "2006-07",
            "to": "2012-02",
            "occupation": "OWNER-GENERAL-MANAGER",
            "company": "SHANGHAI CHONGYUAN MEDICAL DEVICE CO.,LTD",
            "city": "SHANGHAI",
            "country": "China",
            #'province': '',
        },
    ],
}

test_5709 = {
    ## ------- header -------
    "uci": "93379194",
    "service_language": "English",
    "apply_extend_student": True,
    "restore_student": False,
    "new_permit": False,
    ## ------ personal details ------
    "last_name": "Wu",
    "first_name": "Yifan",
    "used_first_name": "Bufan",
    "used_last_name": "Wu",
    "sex": "Male",
    "dob": "1997-10-08",
    "city_of_birth": "Suzhou",
    "country_of_birth": "China",
    "citizenship": "China",
    "current_residence": {
        "country": "Canada",
        "status": "Student",
        "other": "",
        "from": "2019-05-19",
        "to": "2022-08-29",
    },
    #'previous_residence': [],
    #'applying_country': {},
    "current_mariage": {
        "status": "Common-Law",
        "married_date": "2021-01-01",
        "spouse_last_name": "Li",
        "spouse_first_name": "Xiaolu",
        "spouse_is_canadian": False,
    },
    # "previous_mariage": {},
    ## ------ language ------
    "native_language": "Mandarin",
    "english_french": "English",
    #'preferred_language': '',
    "language_test": "Yes",
    ## ------ passport ------
    "passport": {
        "number": "E18826275",
        "country_of_issue": "China",
        "issue_date": "2014-06-09",
        "expiry_date": "2024-06-08",
        #'taiwan_issued': '',
        #'israel_issued': '',
    },
    ## ------ national ID ------
    "national_id": {
        "number": "32050319971008001x",
        "country_of_issue": "CHN (China)",
        "issue_date": "2018-07-12",
        "expiry_date": "2028-07-12",
    },
    ## ------ US PR card ------
    # "us_pr_card": {
    #     "number": "",
    #     "expiry_date": "",
    # },
    ## ------ contact information ------
    "mailing_address": {
        "po_box": "",
        "apt": "302",
        "street_no": "2048",
        "street_name": "W 41st Ave",
        "city": "Vancouver",
        "country": "Canada",
        "province": "AB",
        "postal_code": "V6M 1Y8",
    },
    #'residential_address': {},
    "telephone_no": {
        "canada": "Yes",
        "type": "Cellular",
        "country_code": "1",
        "number": "6047820166",
        "ext": "",
    },
    #'alternate_no': {},
    #'fax_no': {},
    "email": "wyf19971008@gmail.com",
    ## ------ coming into Canada ------
    "coming_into_canada": {
        "date": "2015-12-23",
        "place": "Vancouver",
        "purpose": "Study",
        #'other': '',
        "last_date": "2019-05-19",
        "last_place": "Vancouver",
        "last_document_number": "F313325360",
    },
    ## ------ details of study in Canada ------
    "study_detail": {
        "school": "Hanson College",
        "level": "College - Diploma",
        "field": "Business/Commerce",
        # CAUSION!!! the city list not 100 percent match with option_list cities
        "address_of_school": {
            "province": "BC",
            "city": "New Westminster",
            "address": "206-960 Quayside Drive",
        },
        "institution": "O19618862812",
        "student_id": "H10003244",
        "from": "2022-09-01",
        "to": "2023-12-31",
        "cost": {
            "tuition": "5000",
            "room": "10000",
            "other": "",
        },
        "funds": "30000",
        "expense_by": "Myself",
        #'other': '',
        "apply_for_work": False,
        #'work_permit_type': '',
        # "caq": {
        #    "cert_number": "",
        #    "expiry_date": "",
        # },
    },
    ## ------ education ------
    "education": {
        "from": "2019-09",
        "to": "2021-12",
        "field": "Business,diploma",
        "school": "Hanson College (Cambrian College New Westminster)",
        "city": "New Westminster",
        "country": "Canada",
        "province": "BC",
    },
    ## ------ Employment ------
    "employment": [
        {
            "from": "2020-09",
            "to": "2021-12",
            "occupation": "Part-time working /Office Assistant",
            "company": "Beeta International Education Group",
            "city": "Vancouver",
            "country": "Canada",
            "province": "BC",
        },
    ],
}

test_5710 = {
    ## ------- header -------
    "uci": "93390605",
    "service_language": "English",
    "apply_same_employer": True,
    "apply_new_employer": False,
    "restore_worker": False,
    "new_permit": False,
    ## ------ personal details ------
    "last_name": "Li",
    "first_name": "Haiqiao",
    # "used_first_name": "",
    # "used_last_name": "",
    "sex": "Female",
    "dob": "1971-01-10",
    "city_of_birth": "Xi an, Shannxi",
    "country_of_birth": "China",
    "citizenship": "China",
    "current_residence": {
        "country": "Canada",
        "status": "Worker",
        "other": "",
        "from": "2018-12-28",
        "to": "2022-08-04",
    },
    #'previous_residence': [],
    #'applying_country': {},
    "current_mariage": {
        "status": "Married",
        "married_date": "1999-12-23",
        "spouse_last_name": "Li",
        "spouse_first_name": "Quhu",
        "spouse_is_canadian": False,
    },
    # "previous_mariage": {},
    ## ------ language ------
    "native_language": "Mandarin",
    "english_french": "Neither",
    #'preferred_language': '',
    "language_test": "No",
    ## ------ passport ------
    "passport": {
        "number": "E12745538",
        "country_of_issue": "China",
        "issue_date": "2013-02-05",
        "expiry_date": "2023-02-04",
        #'taiwan_issued': '',
        #'israel_issued': '',
    },
    ## ------ national ID ------
    "national_id": {
        "number": "610103197101100423",
        "country_of_issue": "CHN (China)",
        "issue_date": "2014-04-17",
        "expiry_date": "2034-04-17",
    },
    ## ------ US PR card ------
    # "us_pr_card": {
    #     "number": "",
    #     "expiry_date": "",
    # },
    ## ------ contact information ------
    "mailing_address": {
        "po_box": "",
        "apt": "",
        "street_no": "2963",
        "street_name": "Constellation Ave",
        "city": "Victoria",
        "country": "Canada",
        "province": "AB",
        "postal_code": "V9B 0L9",
    },
    #'residential_address': {},
    "telephone_no": {
        "canada": "Yes",
        "type": "Cellular",
        "country_code": "1",
        "number": "2505802709",
        "ext": "",
    },
    "alternate_no": {
        "canada": "Yes",
        "type": "Business",
        "country_code": "1",
        "number": "2507482575",
        "ext": "",
    },
    "fax_no": {
        "canada": "Yes",
        "country_code": "1",
        "number": "2507483562",
        "ext": "",
    },
    "email": "750024046@qq.com",
    ## ------ coming into Canada ------
    "coming_into_canada": {
        "date": "2018-12-28",
        "place": "Pacific Highway,BC",
        "purpose": "Work",
        #'other': '',
        "last_date": "2019-09-12",
        "last_place": "Vancouver, BC",
        "last_document_number": "U512109881",
    },
    ## ------ details of intended work in Canada ------
    "work_detail": {
        "type": "Exemption from labour Market Impact Assessment",
        # "other": "",
        "employer_name": "CENTRAL GLASS(DUNCAN)LTD.",
        "employer_address": "2856 ROBERTS ROAD, DUNCAN, BC V9L 6W3 CANADA",
        "province": "BC",
        "city": "Duncan",
        "address": "2856 ROBERTS ROAD",
        "job_title": "DIRECTOR OF OPERATIONS",
        "description": "DAILY OPERATIONS INCLUDING PURCHASING,PRODUCTION,QUALITY CONTROL,STAFFING, FINANCING,ETC.",
        "from": "2022-09-01",
        "to": "2023-02-04",
        "lmia_no": "A0651459",
        # "caq_number": "",
        # "caq_expire": "",
        "pnp": False,
    },
    ## ------ education ------
    "education": {
        "from": "1993-09",
        "to": "1995-07",
        "field": "Decoration, Diploma",
        "school": "Xi'an Academy of Arts",
        "city": "Xi'an",
        "country": "China",
    },
    ## ------ Employment ------
    "employment": [
        {
            "from": "2018-12",
            "to": "2022-01",
            "occupation": "Director of Operations",
            "company": "Central Glass (Duncan), Ltd.",
            "city": "Duncan",
            "country": "Canada",
            "province": "BC",
        },
        {
            "from": "2016-08",
            "to": "2018-07",
            "occupation": "Unemployed, accompany child",
            "company": "N/A",
            "city": "Victoria",
            "country": "Canada",
            "province": "BC",
        },
        {
            "from": "2010-03",
            "to": "2016-07",
            "occupation": "Director of Administration& Artistic Design",
            "company": "Shaanxi Ximei Environmental Art Design Co., Ltd.",
            "city": "Xi'an",
            "country": "China",
        },
    ],
}


def load_form_data(file, form_name):
    model = {
        "1295": M1295Model,
        "5708": M5708Model,
        "5709": M5709Model,
        "5710": M5710Model,
    }
    applicant = model[form_name]([file.name])
    return applicant



import json,copy
from termcolor import colored
from openpyxl import load_workbook, Workbook
from openpyxl.workbook.views import BookView
from openpyxl.utils import get_column_letter
from collections import OrderedDict
from source.infosheet import SheetDict
from source.tablesheet import TableList
# from source.style import style
# from source.format import Format
from source.excelwriter import ExcelWritter

            
class Excel(object):
    """Excel manupulate excel for getting source data, compare excels or sheets data, and write new excel for collecting data.
    
    Args:
        excel_name ([str]): [an excel file name]
        sheets ([list]): [default is None, meaning will get all sheets; If assing a list of sheet names, it will only access the specified sheets.
        sheet name in this Excel class will with its original 'info-' or 'table-' head, but the head will be removed when save to dict for exporting;
        When write to a new excel file, the sheet name of a object of SheetDict will be insert 'info-' in the name, while for TableList object, 'table-' 
        will be inserted too]
    Returns: an Excel object
    """
    def __init__(self,excel_name,sheets=None):
        self.excel_name=excel_name
        self.wb=load_workbook(self.excel_name)
        self.sheetNames=self.wb.sheetnames if sheets==None else sheets    # if not specify sheets, it will handle all sheets
        self.sheet_names=[sheetName for sheetName in self.sheetNames if sheetName.lower().startswith('info-') ] #save info sheet names
        self.table_names=[sheetName for sheetName in self.sheetNames if sheetName.lower().startswith('table-') ]#save table sheet names
        self.sheets=OrderedDict() # save info sheet objects
        self.tables=OrderedDict() # save table sheet objects
        self._getSheetsTables()   # get all info and table sheets objects and save to sheets and tables
    
    def __str__(self):
        return self.excel_name+" with info sheets: "+', '.join(self.sheets)+' , and table sheets: '+' ,'.join(self.tables)

    # = operator. Return True if self and another excel have same sheets, and every sheet has same variables 
    def __eq__(self, another: object):
        if self.sheetNames!=another.sheetNames: return False  # if not all sheet names same, False
        # check if info sheets are same
        pairs=list(zip(self.sheets,another.sheets))
        for self_pair,another_pair in pairs:
            if self_pair!=another.pair: return False
        # check if table sheets are same
        pairs=list(zip(self.tables,another.tables))
        for self_pair,another_pair in pairs:
            if self_pair!=another.pair: return False
        return True 
    
    # + union two excel objs (self | another), and return a new excel obj
    # in excel level, I consider to merge value together with variables. If duplicated, another value will prevail. 
    def __add__(self,another:object):
        # iterate another and insert or replace self's        
        # merge sheets (info and talbe)
        for sheet in another.sheets:
            if sheet not in self.sheets: # sheet is a key of sheets OrderedDict
                self.sheets[sheet]=another.sheets[sheet]
            else:
                self.sheets[sheet]+=another.sheets[sheet]
        for table in another.tables:
            if table not in self.tables: 
                #创建一个新的table
                self.tables[table]=another.tables[table]
            else:
                self.tables[table]+=another.tables[table] # 原来是每个子项都自己算，v1.0.3不做了。 从V1.0.4改回来。
                # append rows to old one; 直接把后面的excel table里面的rows添加到签前面的table里面。
                # self.tables[table].data.extend(another.tables[table].data) # 这个是添加到后面的版本。 只有在V1.0.3里面用。
                
        self.sheet_names+=another.sheet_names
        self.table_names+=another.table_names
        return self
    
    # - set difference, return a new obj with sheets in self but not in another, and with variables in self sheet but not in another if has same sheet
    def __sub__(self,another:object):
        # for sheet in obj.sheets:
        for sheet in self.sheets:
            if sheet in another.sheets: # if sheet in obj and in another, self sheet minus another sheet. sheet is a key of sheets OrderedDict
                self.sheets[sheet]-=another.sheets[sheet]
            # if len(self.sheets[sheet])==0:   # if a sheet is empty, delete it
            #     self.sheets.pop(sheet)
        # table part: for columns sub, only care about variables; while for rows, care about if same row
        # for table in self.tables:
        for table in self.tables:
            if table in another.tables: # if table in obj and in another, self sheet minus another sheet. sheet is a key of sheets OrderedDict
                self.tables[table]-=another.tables[table]
            # if len(self.tables[table])==0:
            #     self.tables.pop(table)
        self.sheet_names=list(set(self.sheet_names)-set(another.sheet_names))
        self.table_names=list(set(self.table_names)-set(another.table_names))
        
        return self 
    
    # copy another's excel's common contents to self
    def copy(self,another:object):
        # iterate another's sheet, replace self's value if it is in self too       
        for sheet in another.sheets:
            if sheet in self.sheets: # sheet is a key of sheets OrderedDict
                self.sheets[sheet]=self.sheets[sheet].copy(another.sheets[sheet])
        for table in another.tables:
            if table in self.tables: 
                self.tables[table]=self.tables[table].copy(another.tables[table]) 
        return self
    """
    Below are methods for getting data from excel files and export 
    """
    #get sheet variable name from sheet name
    def _getVariable(self,talbe_sheet_name):
        if talbe_sheet_name.startswith('info-'):
            return talbe_sheet_name[5:]
        elif talbe_sheet_name.startswith('table-'):
            return talbe_sheet_name[6:]
        else:
            return talbe_sheet_name    
    
    # check if a row is None. Return True if there is at least one column not None
    def _notNone(self,row_data):
        return any([False if x==None else True  for x in row_data])
    
    # get info sheet object
    def _getSheet(self,sheet):
        values=list(self.wb[sheet].values)
        values=[x for x in values if self._notNone(x)]  #clean data by removing None 
        sheet_obj=SheetDict(values)
        setattr(sheet_obj,'name',self._getVariable(sheet)) # insert a sheet's name by removing 'info-'
        return sheet_obj
    
    # get table sheet object
    def _getTable(self,table):
        values=list(self.wb[table].values)
        values=[x for x in values if self._notNone(x)]  #clean data by removing None 
        table_obj=TableList(values)
        setattr(table_obj,'name',self._getVariable(table)) # insert a table's name by removing 'table-'
        return table_obj
    
    # get the excel's specified sheets(info and table). In this step, the 'info-' and 'table-' will be removed. The rest sheet name
    # will be used as the key of the object of SheetDict or TableList. 
    def _getSheetsTables(self):
        for sn in self.sheet_names:
            self.sheets[self._getVariable(sn)]=self._getSheet(sn)
        for tn in self.table_names:
            self.tables[self._getVariable(tn)]=self._getTable(tn)
    
    # get sheet object from the dict of self.sheets by specific sheet name,and variable list. default None means all variables. This name is variable name, without info- 
    def getSheet(self,sheet_name,variables=None):
        sheet=self.sheets.get(sheet_name)
        if sheet:
            if variables:
                data={k:v for k,v in sheet.data.items() if k in variables}
                sheet.data=data
                return sheet
            else:
                return sheet
        elif sheet==None:
            return None
        elif len(sheet)==0:
            return sheet
        else:
            raise ValueError(f'Sheet info-{sheet_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models')
    
    # get sheet object by specific sheet name. This name is variable name, without table- 
    def getTable(self,table_name,variables=None):
        table=self.tables.get(table_name)
        if table is not None:
            if variables:
                temp_table=copy.deepcopy(table)
                if len(table.data)>0:
                    for index, data in enumerate(table.data):
                        for k in data.__dict__: # check every TableNode object
                            if k not in variables:  # if not in variable list
                                delattr(temp_table.data[index],k)   # delete from returning obj
                temp_table.column_variables=variables   # update the table variables
                temp_table.column_titles=OrderedDict({k:v for k,v in table.column_titles.items() if k in variables}) # update the table titles
                return temp_table                
            else:
                return table
        else:
            raise ValueError(f"Table table-{table_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models")
    
    # get sheet dict by specific sheet name. This name is variable name, without info- 
    def getSheetData(self,sheet_name):
        # sheet_data=OrderedDict()
        if self.sheets.get(sheet_name): 
            sheet=self.sheets.get(sheet_name)
            # get key(variable) and value from object
            sheet_data={k:v.value for k, v in sheet.data.items()}
        elif len(self.sheets.get(sheet_name).data)==0:
            sheet_data={}
        else:
            raise ValueError(f'Sheet info-{sheet_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models')
        
        return sheet_data
    
    # get table list by specific sheet name. This name is variable name, without table- 
    def getTableData(self,table_name):
        table=self.tables.get(table_name,None)
        # else:
        #     raise ValueError(f"Table table-{table_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models")
        # if table is not existed or data is [], will return []
        if table:
            table_data=[ td.__dict__ for td in table.data ]
            return table_data
        return []
    
    # get all info-sheets and table-sheets data
    @property
    def dict(self):
        the_data={
            **{sheet:self.getSheetData(sheet) for sheet in self.sheets.keys() },
            **{table:self.getTableData(table) for table in self.tables.keys() }    
        }
        return the_data
    
    @property
    def json(self):
        print(json.dumps(self.dict,indent=3,default=str))

    """
    Below is write back to excel  mothods
    """
    def cellStyle(self,cell,format):
        cell.font = format.get('font')
        cell.border = format.get('border')
        cell.fill = format.get('fill')
        cell.alignment = format.get('alignment')
    
    def _formatCells(self,ws,columns):
        for col in range(1,columns+1):
            cell=ws.cell(1,col)
            self.cellStyle(cell,style.get('TITLE_FORMAT'))

    def makeInfoSheet(self,sheet:SheetDict):
        # create a new info sheet in new workbook
        sheet_name='info-'+sheet.name
        new_ws=self.new_wb.create_sheet(sheet_name)
        rows=[]
        rows.append([sheet.sheet_title])
        rows.append(list(sheet.column_titles))
        # not use titled variables 
        # titled_variables=[ x.title() for x in sheet.column_variables]
        rows.append(sheet.column_variables)
        for new_row in sheet.data.values():
            rows.append(list(new_row.__dict__.values()))
        for row in rows:
            new_ws.append(row)
        max_column=get_column_letter(len(sheet.column_titles))
        new_ws.merge_cells(f'A1:{max_column}1')
    
    def makeTableSheet(self,table:TableList):
        # create a new table sheet in new workbook
        table_name='table-'+table.name
        new_ws=self.new_wb.create_sheet(table_name)
        rows=[]
        rows.append([table.sheet_title])
        variables=table.column_variables
        titles=len(variables)>0 and [ table.column_titles.get(v) for v in variables]
        rows.append(titles)
        rows.append(variables)
        for new_row in table.data:
            rows.append(list(new_row.__dict__.values()))
        for row in rows:
            row and new_ws.append(row)
        max_column=get_column_letter(len(table.column_titles))
        new_ws.merge_cells(f'A1:{max_column}1')
            
    def makeExcel(self,file_name,protection=False,sheets=None,tables=None):
        # 创建新的excel文件
        self.new_wb=Workbook()
        
        self.new_wb.views=[BookView(xWindow=0, yWindow=0, windowWidth=18140, windowHeight=15540)]
        
        # handle info sheets. if None, including all sheets/tables, if [], including nothing. if has sheets/tables, including those sheets/tables
        sheets=sheets if sheets is not None else self.sheets.values()
        tables=tables if tables is not None else self.tables.values()
        [self.makeInfoSheet(sheet) for sheet in sheets if sheet is not None]
        # handle table sheets
        [self.makeTableSheet(table) for table in tables if table is not None]
        #保存为新文件
        if len(self.sheets)>0 or len(self.tables)>0:    
            self.new_wb.remove(self.new_wb['Sheet']) #exists only when niether info nor table has sheet 
        # old version, removed
        # f=Format(self.new_wb)
        # f.setAll()
        # if protection:
        #     f.protectAll()
        self.new_wb.save(file_name) #先保存
        self.formatExcel(file_name,protection) # format这个文件
        
    def formatExcel(self,output_file_name,protection=True):
        input_excel_obj=Excel(output_file_name) #get excel obj for format
        ew = ExcelWritter(input_excel_obj,output_file_name)
        if protection:
            password=input("please input password:")
            ew.create(protection, password)
        else:
            ew.create(protection=False)
        




valid_programs=['5257','1294','1295','5708','5709','5710','0104','5645']

class ModelImporter(object):
    def __init__(self,program_code) -> None:
        self.program_code=program_code
        
    def loadModel(self,model_type):
        module_prefix="m" if model_type=="DataModel" else "f"
        model_prefix="M" if model_type=="DataModel" else "F"
        module_path=f"model.tr.{module_prefix}{self.program_code}"
        class_list=[f"{model_prefix}{self.program_code}Model"]
        the_class=__import__(module_path,fromlist=class_list)
        return getattr(the_class,class_list[0])
    
class ExcelMaker(ModelImporter):
    def makeExcelBasedOnModel(self,output_excel):
        # 获取数据模型
        data_model=self.loadModel("DataModel")
        data_model(output_excel_file=output_excel)
        return output_excel

class XmlReader(ModelImporter):
    def readXmlData2Excel(self, xml_file,output_excel):
        #获取pdf表格模型
        form_model=self.loadModel('FormModel')
        form=form_model(xml_file)
        form._makeExcel(output_excel,protection=True)




from audioop import add
from datetime import datetime
from webform.models.definition import Action, Role
from webform.models.prmodel import PrModel
from model.common.address import Addresses
from typing import List
from webform.models.prmodel import Family


class F5669:
    # dict of field ids
    fild_id = {
        "pa": {
            "last_name": "#familyName",
            "first_name": "#givenName",
            "native_fullname": "#nativeFullName",
            "dob": "#dob",
        },
        "dad": {
            "last_name": "#familyNameFather",
            "first_name": "#givenNameFather",
            "dob": "#sectionAFormFatherDOB",
            "death_date": "#sectionAFormFatherDeceasedDate",
            "city": "#sectionAFormFatherCityOfBirth",
            "country": "#sectionAFormFatherCountryOfBirth",
        },
        "mom": {
            "last_name": "#familyNameMother",
            "first_name": "#givenNameMother",
            "dob": "#sectionAFormMotherDOB",
            "death_date": "#sectionAFormMotherDeceasedDate",
            "city": "#sectionAFormMotherCityOfBirth",
            "country": "#sectionAFormMotherCountryOfBirth",
        },
    }

    def __init__(self, pa: PrModel, sp: PrModel | None, dps: List[PrModel]):
        self.pa = pa
        self.sp = sp
        self.dps = dps

    @property
    def pdf(self):
        return [{"action_type": Action.Pdf.value}]

    def goTo5669(self):
        return [
            {
                "action_type": "WebPage",
                "page_name": "Go to Imm5669 ",
                "actions": [],
                "next_page_tag":"body > pra-root > pra-localized-app > main > div > pra-imm5669 > pra-imm5669-start > pra-family-table > button",
                "id": "body > pra-root > pra-localized-app > main > div > pra-intake-landing-page > pra-web-form-table > div > table > tbody > tr:nth-child(25) > td:nth-child(5) > button",
            }
        ]

    def add_family_member(self):
        return [
            {
                "action_type": "WebPage",
                "page_name": "Add new family member",
                "actions": [],
                "next_page_tag":"#sectionAFormMotherCountryOfBirth",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > pra-imm5669-start > pra-family-table > button",
            }
        ]

    def pick_whom(self, role):
        return [
            {
                "action_type": Action.Radio.value,
                "label": "Indicate whether you are filling out this form for",
                "id": "#principalApplicant" if role == Role.PA else "#principalOther",
            }
        ]

    def back2Application(self):
        return [
            {
                "action_type": "WebPage",
                "page_name": "Complete and return to application",
                "actions": [],
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-secondary",
            }
        ]

    def sectionABlockApplicant(self, person: PrModel):
        return [
            {
                "action_type": "Input",
                "label": "Family name",
                "value": person.personal.last_name,
                "id": F5669.fild_id["pa"]["last_name"],
                "length": 57,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Given name",
                "value": person.personal.first_name,
                "id": F5669.fild_id["pa"]["first_name"],
                "length": 57,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Full name",
                "value": person.personal.native_first_name
                + " "
                + person.personal.native_last_name,
                "id": F5669.fild_id["pa"]["native_fullname"],
                "length": 90,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Birth information",
                "value": person.personal.dob.strftime("%Y/%m/%d"),
                "id": F5669.fild_id["pa"]["dob"],
                "length": 10,
                "required": True,
            },
        ]

    def sectionABlock(self, relationship, family_member: Family):
        return [
            {
                "action_type": "Input",
                "label": "Family name",
                "value": family_member.last_name+"  "+family_member.native_last_name,
                "id": F5669.fild_id[relationship]["last_name"],
                "length": 57,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Given name",
                "value": family_member.first_name+"  "+family_member.native_first_name,
                "id": F5669.fild_id[relationship]["first_name"],
                "length": 57,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Date of birth",
                "value": family_member.date_of_birth.strftime("%Y/%m/%d"),
                "id": F5669.fild_id[relationship]["dob"],
                "length": 10,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Date of death",
                "value": family_member.date_of_death.strftime("%Y/%m/%d")
                if family_member.date_of_death
                else None,
                "id": F5669.fild_id[relationship]["death_date"],
                "length": 10,
                "required": False,
            },
            {
                "action_type": "Input",
                "label": "Town",
                "value": family_member.place_of_birth,
                "id": F5669.fild_id[relationship]["city"],
                "length": 31,
                "required": True,
            },
            {
                "action_type": "Input",
                "label": "Town",
                "value": family_member.birth_country,
                "id": F5669.fild_id[relationship]["country"],
                "length": 31,
                "required": True,
            },
        ]

    def makeSectionB(self, person: PrModel):
        return [
            {
                "action_type": Action.Radio.value,
                "id": "#isConvictedInCanada_yes"
                if person.prbackground.q1
                else "#isConvictedInCanada_no",
                "label": "Yes" if person.prbackground.q1 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isConvictedOutsideCanada_yes"
                if person.prbackground.q2
                else "#isConvictedOutsideCanada_no",
                "label": "Yes" if person.prbackground.q2 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isClaimedRefugeeProtection_yes"
                if person.prbackground.q3
                else "#isClaimedRefugeeProtection_no",
                "label": "Yes" if person.prbackground.q3 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isConvictedInCanada_yes"
                if person.prbackground.q3
                else "#isConvictedInCanada_no",
                "label": "Yes" if person.prbackground.q3 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isRefusedRefugeeOrVisa_yes"
                if person.prbackground.q4
                else "#isRefusedRefugeeOrVisa_no",
                "label": "Yes" if person.prbackground.q4 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isOrderedToLeaveCountry_yes"
                if person.prbackground.q5
                else "#isOrderedToLeaveCountry_no",
                "label": "Yes" if person.prbackground.q5 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isWarCriminal_yes"
                if person.prbackground.q6
                else "#isWarCriminal_no",
                "label": "Yes" if person.prbackground.q6 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isCommittedActOfViolence_yes"
                if person.prbackground.q7
                else "#isCommittedActOfViolence_no",
                "label": "Yes" if person.prbackground.q7 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isAssociatedWithViolentGroup_yes"
                if person.prbackground.q8
                else "#isAssociatedWithViolentGroup_no",
                "label": "Yes" if person.prbackground.q8 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isMemberOfCriminalOrg_yes"
                if person.prbackground.q9
                else "#isMemberOfCriminalOrg_no",
                "label": "Yes" if person.prbackground.q9 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isDetainedOrJailed_yes"
                if person.prbackground.q10
                else "#isDetainedOrJailed_no",
                "label": "Yes" if person.prbackground.q10 else "No",
            },
            {
                "action_type": Action.Radio.value,
                "id": "#isPhysicalOrMentalDisorder_yes"
                if person.prbackground.q11
                else "#isPhysicalOrMentalDisorder_no",
                "label": "Yes" if person.prbackground.q11 else "No",
            },
            {
                "action_type": Action.Input.value,
                "id": "#additionalDetails",
                "label": "Explanation",
                "value": person.prbackground.details,
                "length": 679,
                "required": False,
            },
        ]

    def makeSectionC(self, person: PrModel):
        # education summary
        edu_summary = [
            {
                "action_type": Action.Input.value,
                "label": "Primary school years",
                "id": "#elementarySchoolYears",
                "value": str(person.personal.primary_school_years),
                "length": 2,
                "required": True,
            },
            {
                "action_type": Action.Input.value,
                "label": "Secondary school years",
                "id": "#secondarySchoolYears",
                "value": str(person.personal.secondary_school_years),
                "length": 2,
                "required": True,
            },
            {
                "action_type": Action.Input.value,
                "label": "University school years",
                "id": "#universityAndCollegeYears",
                "value": str(person.personal.post_secondary_school_years),
                "length": 2,
                "required": True,
            },
            {
                "action_type": Action.Input.value,
                "label": "Other school years",
                "id": "#otherSchoolYears",
                "value": str(person.personal.other_school_years),
                "length": 2,
                "required": True,
            },
        ]

        # loop each educaton
        value = []
        for index, edu in enumerate(person.education):
            if edu.end_date=='Present':edu.end_date=None
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from" + str(index),
                    "value": edu.start_date.strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#to" + str(index),
                    "value": edu.end_date.strftime("%Y/%m") if edu.end_date else None,
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "School name",
                    "id": "#nameOfInstitution" + str(index),
                    "value": edu.school_name,
                    "length": 24,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City and country",
                    "id": "#cityAndCountry" + str(index),
                    "value": edu.city + "/" + edu.country
                    if edu.city and edu.country
                    else None,
                    "length": 21,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Type of certificate",
                    "id": "#typeOfDiploma" + str(index),
                    "value": edu.education_level,
                    "length": 17,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Field of study",
                    "id": "#fieldOfStudy" + str(index),
                    "value": edu.field_of_study,
                    "length": 17,
                    "required": True,
                },
            ]
            value.append(block)

        return edu_summary + [
            {
                "action_type": Action.RepeatSection.value,
                "button_text": "Add another",
                "value": value,
            }
        ]

    def makeSectionD(self, person: PrModel):
        # loop each personal history
        value = []
        for index, history in enumerate(person.history):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from" + str(index),
                    "value": history.start_date.strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#to" + str(index),
                    "value": history.end_date.strftime("%Y/%m")
                    if history.end_date
                    else datetime.today().strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Activity",
                    "id": "#activity" + str(index),
                    "value": history.activity,
                    "length": 21,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City and country",
                    "id": "#cityAndCountry" + str(index),
                    "value": history.city_and_country,
                    "length": 62,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Status",
                    "id": "#status" + str(index),
                    "value": history.status,
                    "length": 14,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Name of employer or school",
                    "id": "#nameOfEmployerOrSchool" + str(index),
                    "value": history.name_of_company_or_school,
                    "length": 27,
                    "required": True,
                },
            ]
            value.append(block)

        return [
            {
                "action_type": Action.RepeatSection.value,
                "button_text": "Add another",
                "value": value,
            }
        ]

    def makeSectionE(self, person: PrModel):
        # loop each political member history
        value = []
        for index, member in enumerate(person.member):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from" + str(index),
                    "value": member.start_date.strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#to" + str(index),
                    "value": member.end_date.strftime("%Y/%m")
                    if member.end_date
                    else datetime.today().strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Name of organization",
                    "id": "#nameOfOrganization" + str(index),
                    "value": member.organization_name,
                    "length": 17,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Type of organization",
                    "id": "#typeOfOrganization" + str(index),
                    "value": member.organization_type,
                    "length": 20,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Activity",
                    "id": "#activity" + str(index),
                    "value": member.position,
                    "length": 26,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City and country",
                    "id": "#cityAndCountry" + str(index),
                    "value": member.city + "/" + member.country,
                    "length": 17,
                    "required": True,
                },
            ]
            value.append(block)

        return (
            [
                {
                    "action_type": Action.RepeatSection.value,
                    "button_text": "Add another",
                    "value": value,
                }
            ]
            if len(person.member) > 0
            else [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from0",
                    "value": "None",
                    "length": 7,
                    "required": True,
                }
            ]
        )

    def makeSectionF(self, person: PrModel):
        # loop each government position
        value = []
        for index, government in enumerate(person.government):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#dateFrom" + str(index),
                    "value": government.start_date.strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#to" + str(index),
                    "value": government.end_date.strftime("%Y/%m")
                    if government.end_date
                    else datetime.today().strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Country",
                    "id": "#cityAndCountry" + str(index),
                    "value": government.country,
                    "length": 29,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Department",
                    "id": "#department" + str(index),
                    "value": government.department,
                    "length": 30,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Activity",
                    "id": "#activity" + str(index),
                    "value": government.position,
                    "length": 21,
                    "required": True,
                },
            ]
            value.append(block)

        return (
            [
                {
                    "action_type": Action.RepeatSection.value,
                    "button_text": "Add another",
                    "value": value,
                }
            ]
            if len(person.government) > 0
            else [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#dateFrom0",
                    "value": "None",
                    "length": 7,
                    "required": True,
                }
            ]
        )

    def makeSectionG(self, person: PrModel):
        # loop each military service
        value = []
        for index, military in enumerate(person.military):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "Country",
                    "id": "#country" + str(index),
                    "value": military.country,
                    "length": 30,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Branch",
                    "id": "#branchOfService" + str(index),
                    "value": military.service_detail,
                    "length": 64,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from" + str(index),
                    "value": military.start_date.strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#to" + str(index),
                    "value": military.end_date.strftime("%Y/%m")
                    if military.end_date
                    else datetime.today().strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Rank",
                    "id": "#rank" + str(index),
                    "value": military.rank,
                    "length": 14,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Reaons for end of service",
                    "id": "#reasonsEndService" + str(index),
                    "value": military.rank,
                    "length": 200,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Combat details",
                    "id": "#combatDetails" + str(index),
                    "value": military.combat_detail,
                    "length": 679,
                    "required": True,
                },
            ]
            value.append(block)

        return (
            [
                {
                    "action_type": Action.RepeatSection.value,
                    "button_text": "Add another",
                    "value": value,
                }
            ]
            if len(person.military) > 0
            else [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from0",
                    "value": "None",
                    "length": 7,
                    "required": True,
                }
            ]
        )

    def makeSectionH(self, person: PrModel):
        # loop each military service
        value = []
        for index, address in enumerate(person.addresshistory):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#from" + str(index),
                    "value": address.start_date.strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#to" + str(index),
                    "value": address.end_date.strftime("%Y/%m")
                    if address.end_date
                    else datetime.today().strftime("%Y/%m"),
                    "length": 7,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Street and number",
                    "id": "#street" + str(index),
                    "value": address.street_and_number,
                    "length": 30,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City",
                    "id": "#city" + str(index),
                    "value": address.city,
                    "length": 14,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Province",
                    "id": "#provinceOrState" + str(index),
                    "value": address.province,
                    "length": 14,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Country",
                    "id": "#country" + str(index),
                    "value": address.country,
                    "length": 14,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Post code",
                    "id": "#postalCode" + str(index),
                    "value": address.post_code,
                    "length": 9,
                    "required": True,
                },
            ]
            value.append(block)

        return [
            {
                "action_type": Action.RepeatSection.value,
                "button_text": "Add another",
                "value": value,
            }
        ]

    def sectionA(self, person: PrModel, role):
        applicant = self.sectionABlockApplicant(person)

        father_obj = [
            person
            for person in person.family
            if person.relationship.lower() == "father"
        ]
        mother_obj = [
            person
            for person in person.family
            if person.relationship.lower() == "mother"
        ]
        father = self.sectionABlock("dad", father_obj[0]) if father_obj else []
        mother = self.sectionABlock("mom", mother_obj[0]) if mother_obj else []
        whom = []
        match role:
            case Role.PA:
                whom = self.pick_whom(role=Role.PA)
            case Role.SP:
                whom = self.pick_whom(role=Role.SP)
            case Role.DP:
                whom = self.pick_whom(role=Role.DP)

        actions = whom + applicant + father + mother + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section A: Personal details",
                "actions": actions,
                "next_page_tag":"#isPhysicalOrMentalDisorder_no",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionB(self, person: PrModel):
        actions = self.makeSectionB(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section B: Questionarie",
                "actions": actions,
                "next_page_tag":"#fieldOfStudy0",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionC(self, person: PrModel):
        actions = self.makeSectionC(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section C: Education",
                "actions": actions,
                "next_page_tag":"#nameOfEmployerOrSchool0",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionD(self, person: PrModel):
        actions = self.makeSectionD(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section D: Personal history",
                "actions": actions,
                "next_page_tag":"#activities0",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionE(self, person: PrModel):
        actions = self.makeSectionE(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section E: Membership and association with organization",
                "actions": actions,
                "next_page_tag":"#department0",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionF(self, person: PrModel):
        actions = self.makeSectionF(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section F: Government positions",
                "actions": actions,
                "next_page_tag":"#combatDetails0",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionG(self, person: PrModel):
        actions = self.makeSectionG(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section G: Military and paramilitary service",
                "actions": actions,
                "next_page_tag":"#postalCode0",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def sectionH(self, person: PrModel):
        actions = self.makeSectionH(person) + self.pdf
        return [
            {
                "action_type": "WebPage",
                "page_name": "Section H: Addresses",
                "actions": actions,
                "next_page_tag":"body > pra-root > pra-localized-app > main > div > pra-imm5669 > pra-imm5669-start > pra-family-table > button",
                "id": "body > pra-root > pra-localized-app > main > div > pra-imm5669 > lib-navigation-buttons > div > button.btn.btn-primary",
            }
        ]

    def fill(self):
        pa_form = (
            self.add_family_member()
            + self.sectionA(self.pa, role=Role.PA)
            + self.sectionB(self.pa)
            + self.sectionC(self.pa)
            + self.sectionD(self.pa)
            + self.sectionE(self.pa)
            + self.sectionF(self.pa)
            + self.sectionG(self.pa)
            + self.sectionH(self.pa)
        )
        sp_form = (
            self.add_family_member()
            + self.sectionA(self.sp, role=Role.SP)
            + self.sectionB(self.sp)
            + self.sectionC(self.sp)
            + self.sectionD(self.sp)
            + self.sectionE(self.sp)
            + self.sectionF(self.sp)
            + self.sectionG(self.sp)
            + self.sectionH(self.sp)
            if self.sp
            else []
        )
        dp_forms = []
        for dp in self.dps:
            if dp.personal.age >= 18:
                dp_forms += (
                    self.add_family_member()
                    + self.sectionA(dp, role=Role.DP)
                    + self.sectionB(dp)
                    + self.sectionC(dp)
                    + self.sectionD(dp)
                    + self.sectionE(dp)
                    + self.sectionF(dp)
                    + self.sectionG(dp)
                    + self.sectionH(dp)
                )

        return self.goTo5669() + pa_form + sp_form + dp_forms + self.back2Application()



# import context
import argparse,os
from termcolor import colored
from model.experience.resume import ResumeModel,ResumeModelDocxAdapater
from model.experience.ec import EmploymentModel,EmploymentModelDocxAdapater

def creatExcel(model,excel_file):
    model(output_excel_file=excel_file)
    print(colored(f'{excel_file} is not existed, so we created the excel file based on your data structure.','green'))
    print(colored('Please fill the excel with data and do it again','yellow'))
    
def generateDocx(source_excel,model,adapter,outpout,template_num=None):
    context=model(excels=[source_excel])
    context=adapter(context)
    context.make(outpout,template_num) if template_num else  context.make(outpout)

models={
    "rs":{"model":ResumeModel,'adaptor':ResumeModelDocxAdapater},
    "ec":{"model":EmploymentModel,'adaptor':EmploymentModelDocxAdapater},
} 

    
def main():
    """ arguments:
        -e: excel file as source
        -t: output word file name
        -d: --> rs: resume, ec: employment certificates
        -tn: resume template number
    """
    parser=argparse.ArgumentParser(description="used for processing lmia application")
    # parser.add_argument("-s", "--stream", help="input stream code. 5593 for pr, 5626 for HWS or 5627 for LWS ")
    parser.add_argument("-e", "--excel", help="input excel file name including the data for your specific stream")
    parser.add_argument("-t", "--to", help="input docx file name for output")
    parser.add_argument("-d", "--document", help="input which kind of document to generate. type list: rs: resume, ec: employment certificates")
    parser.add_argument("-tn", "--template_no", help="resume template number (1-5)")
    args = parser.parse_args()

    
    if args.excel and args.to and args.document:
        try: 
            model=models[args.document]['model']
            adaptor=models[args.document]['adaptor']
        except KeyError as e:
            print(colored(f"{e} is not a valid document type. Valid type list is: rs, ec",'red'))
            return 
        # if input excel file, then generate it
        if not os.path.isfile(args.excel):
            if args.document!='ec': 
                creatExcel(model,args.excel)
            else:
                print(colored(f'{args.excel} is not existed.','red'))
                print(colored('You can use personal excel pa.xlsx and fill the employment rows, and then try again...','blue'))
            return 
        generateDocx(args.excel,model,adaptor,args.to,args.template_no)
    
if __name__=='__main__':
    main()






import argparse
from termcolor import colored
from utils import utils

# 1. IELTS->CLB level
# CLB Level:	Reading	Writing	Listening	Speaking
clb_ielts_table = [
    [10, 8.0, 7.5, 8.5, 7.5],
    [9, 7.0, 7.0, 8.0, 7.0],
    [8, 6.5, 6.5, 7.5, 6.5],
    [7, 6.0, 6.0, 6.0, 6.0],
    [6, 5.0, 5.5, 5.5, 5.5],
    [5, 4.0, 5.0, 5.0, 5.0],
    [4, 3.5, 4.0, 4.5, 4.0]
]

# 输入CLB值 返回对应的雅思r w l s分数列表
def clb2ielts(clb):
    clb=10 if clb>10 else clb
    if clb<4:
        return [0,0,0,0]
     
    index=10-clb
    return clb_ielts_table[index][1:5]

# 输入一组雅思成绩list：r w l s, 输出一个list对应CLB level： r w l s
def ielts_2_clbs(ielts):
    clb_level = []
    for column in range(1, 5):  # 从读到说，4个列，从1开始，开始找对应的clb级别
        factor = float(ielts[column - 1])  # r w l s每个子项
        for level in range(0, 7):  # 从CLB 级别10一直找到4 7行
            if factor > clb_ielts_table[0][column]:  # 如果语言成绩高于最高分，按照最高分算
                clb_level.append(10) 
                break
            if factor < clb_ielts_table[6][column]:  # 如果低于最低分，则CLB Level =0 跳出本列循环
                clb_level.append(0)
                break
            # 要算每个子项的CLB值
            if clb_ielts_table[level][column] == ielts[column - 1]:
                clb_level.append(clb_ielts_table[level][0])  # 得到该雅思成绩对应的CLB 级别
                break
        # 如果某项分数不在阵列中
        for level in range(1, 7):
            if clb_ielts_table[level][column] < factor < clb_ielts_table[level - 1][column]:
                clb_level.append(clb_ielts_table[level][0])  # 得到该雅思成绩对应的CLB 级别
                break

    return clb_level


# 输入一组雅思成绩list：r w l s, 输出一个CLB level 数字
def ielts_2_clb(ielts):
    clb = 0
    for row in clb_ielts_table:
        if (float(ielts[0]) >= row[1]) & (float(ielts[1]) >= row[2]) & (float(ielts[2]) >= row[3]) & (float(ielts[3]) >= row[4]):
            clb = row[0]
            break
    return clb


# 2 CLPIP和CLB是一一对应关系。
# Canadian English Language Proficiency Index Program (CELPIP) – General Test score equivalency chart
# CLB Level	Reading	Writing	Listening	Speaking
# 10	10	10	10	10
# 9	9	9	9	9
# 8	8	8	8	8
# 7	7	7	7	7
# 6	6	6	6	6
# 5	5	5	5	5
# 4	4	4	4	4

def celpip_2_clb(rwls):
    has_error = False
    for s in rwls:
        if not isinstance(s, int):
            print("the CELPIP must be integer")
            has_error = True
    if not has_error:
        if min(rwls)>10:
            result=10 
        elif min(rwls)<0:
            result=0
        else:
            result=min(rwls)
        return result
    else:
        return 0


def celpip_2_clbs(rwls):
    has_error = False
    for s in rwls:
        if not isinstance(s, int):
            print("the CELPIP must be integer")
            has_error = True
    if not has_error:
        i=0
        for x in rwls:
            if x>10:
                rwls[i]=10
            elif x<0:
                rwls[i]=0
            else:
                rwls[i]=x
            i+=1
        return rwls
    else:
        return []


# 3 TEF
# Test d’évaluation de français pour le Canada (TEF Canada)
# Test score equivalency chart
# CLB Level	Reading	Writing	Listening	Speaking
# 10	    263-277	393-415	316-333	    393-415
# 9	        248-262	371-392	298-315	    371-392
# 8	        233-247	349-370	280-297	    349-370
# 7	        207-232	310-348	249-279	    310-348
# 6	        181-206	271-309	217-248	    271-309
# 5	        151-180	226-270	181-216	    226-270
# 4	        121-150	181-225	145-180	    181-225

tef_clb_table = [
    [10, 263, 277, 393, 415, 316, 333, 393, 415],
    [9, 248, 262, 371, 392, 298, 315, 371, 392],
    [8, 233, 247, 349, 370, 280, 297, 349, 370],
    [7, 207, 232, 310, 348, 249, 279, 310, 348],
    [6, 181, 206, 271, 309, 217, 248, 271, 309],
    [5, 151, 180, 226, 270, 181, 216, 226, 270],
    [4, 121, 150, 181, 225, 145, 180, 181, 225]
]


def tef_2_clb(rwls):
    has_error = False
    for s in rwls:
        if not isinstance(s, int):
            print("the TEF must be integer")
            has_error = True
    if not has_error:

        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(tef_clb_table[row][start], tef_clb_table[row][end] + 1):
                    clb.append(tef_clb_table[row][0])
                    break
        return min(clb)
    else:
        return 0

def tef_2_clbs(rwls):
    has_error = False
    for s in rwls:
        if not isinstance(s, int):
            print("the TEF must be integer")
            has_error = True
    if rwls[0] not in range(121,278):
        print('read score is invalid')
        exit(1)
    if rwls[1] not in range(181,416):
        print('write score is invalid')
        exit(1)
    if rwls[2] not in range(145,334):
        print('listen score is invalid')
        exit(1)
    if rwls[3] not in range(181,416):
        print('write score is invalid')
        exit(1)
    
    if not has_error:
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(tef_clb_table[row][start], tef_clb_table[row][end] + 1):
                    clb.append(tef_clb_table[row][0])
                    break
        return clb
    else:
        return 0

# 输入CLB值 返回对应的tef r w l s分数列表
def clb2tef(clb):
    clb=10 if clb>10 else clb
    if clb<4:
        return [0,0,0,0,0,0,0,0]
     
    index=10-clb
    return tef_clb_table[index][1:9]

# 4 TCF
# Test de connaissance du français pour le Canada (TCF Canada)
# Test score equivalency chart
# CLB Level	    Reading	Writing	Listening	Speaking
# 10 and above	549-699	16-20	549-699	16-20
# 9	            524-548	14-15	52        
# 8	            499-523	12-13	503-522	12-13
# 7	            453-498	10-11	458-502	10-11
# 6	            406-452	7-9	    398-457	7-9
# 5	            375-405	6	    369-397	6
# 4	            342-374	4-5	    331-368	4-5

tcf_clb_table = [
    [10, 549, 699, 16, 20, 549, 699, 16, 20],
    [9, 524, 548, 14, 15, 523, 548, 14, 15],
    [8, 499, 523, 12, 13, 503, 522, 12, 13],
    [7, 453, 498, 10, 11, 458, 502, 10, 11],
    [6, 406, 452, 7, 9, 398, 457, 7, 9],
    [5, 375, 405, 6, 6, 369, 397, 6, 6],
    [4, 342, 374, 4, 5, 331, 368, 4, 5]
]


def tcf_2_clb(rwls):        
    has_error=False
    for s in rwls:
        if not isinstance(s, int):
            print("the TCF must be integer")
            has_error = True
    if rwls[0] not in range(342,700):
        print('read score is invalid')
        exit(1)
    if rwls[1] not in range(4,21):
        print('write score is invalid')
        exit(1)
    if rwls[2] not in range(331,700):
        print('listen score is invalid')
        exit(1)
    if rwls[3] not in range(4,21):
        print('write score is invalid')
        exit(1)
    if not has_error:
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(tcf_clb_table[row][start], tcf_clb_table[row][end] + 1):
                    clb.append(tcf_clb_table[row][0])
                    break
        return min(clb)
    else:
        return 0


def tcf_2_clbs(rwls):
    has_error = False
    for s in rwls:
        if not isinstance(s, int):
            print("the TCF must be integer")
            has_error = True
    if not has_error:
        clb = []
        for column in range(0, 4):  # r w l s
            for row in range(0, 7):
                start = (column + 1) * 2 - 1
                end = (column + 1) * 2
                if rwls[column] in range(tcf_clb_table[row][start], tcf_clb_table[row][end] + 1):
                    clb.append(tcf_clb_table[row][0])
                    break
        return clb
    else:
        return 0

# 输入CLB值 返回对应的tcf r w l s分数列表
def clb2tcf(clb):
    clb=10 if clb>10 else clb
    if clb<4:
        return [0,0,0,0,0,0,0,0]
     
    index=10-clb
    return tcf_clb_table[index][1:9]

def getCLBs(rwls, format):
    format=format.upper()
    if format == 'IELTS':
        clb = ielts_2_clb(rwls)
        clbs = ielts_2_clbs(rwls)
    if format == 'CELPIP':
        clb = celpip_2_clb(rwls)
        clbs = celpip_2_clbs(rwls)
    if format == 'TEF':
        clbs = tef_2_clbs(rwls)        
        clb=tef_2_clb(rwls)
    if format == 'TCF':
        clb = tcf_2_clb(rwls)
        clbs = tcf_2_clbs(rwls)
    return clb, clbs


class Language:
    def __init__(self,r,w,l,s,format):
        self.reading = r
        self.writing = w
        self.listening = l
        self.speaking = s
        self.format = format
        self.clb, [self.clb_r, self.clb_w, self.clb_l, self.clb_s] = getCLBs([self.reading, self.writing, self.listening, self.speaking], self.format)

    def get4(self):
        return self.reading,self.writing,self.listening,self.speaking,self.format        


def main():
    parser=argparse.ArgumentParser(description="Language test and CLB converter")
    # Language
    parser.add_argument("-li", "--level2ielts", help="input clb level")
    parser.add_argument("-lc", "--level2celpip", help="input clb level")
    parser.add_argument("-ltef", "--level2tef", help="input clb level")
    parser.add_argument("-ltcf", "--level2tcf", help="input clb level")

    parser.add_argument("-il", "--ielts2level", help="input ielts 4 score(r w s l)",nargs='+')
    parser.add_argument("-cl", "--celpip2level", help="input celpip 4 score(r w s l)",nargs='+')
    parser.add_argument("-tefl", "--tef2level", help="input tef  4 score(r w s l)",nargs='+')
    parser.add_argument("-tcfl", "--tcf2level", help="input tcf 4 score(r w s l)", nargs='+')

    parser.add_argument("-l", "--level", help="input clb level")
    
    args = parser.parse_args()
    no_variables=True

    if args.level2ielts:
        no_variables=False
        clb=int(args.level2ielts)
        items=[]
        if clb!=None:
            print(colored(f'\nCLB {clb} to IELTS:','green'))
            items.append(['Read','Write','Listen','Speak'])
            items.append(clb2ielts(clb))
            utils.printFList2D(items)
        else:
            print('You should input clb level...')
            exit(1)

    if args.level2celpip:
        no_variables=False
        clb=int(args.level2celpip)
        items=[]
        if clb!=None: 
            if clb>10:
                clb=10
            elif clb<0:
                clb=0
            print(colored(f'\nCLB {clb} to CELPIP:','green'))
            items.append(['Read','Write','Listen','Speak'])
            items.append([clb,clb,clb,clb])
            utils.printFList2D(items)
        else:
            print('You should input clb level...')
            exit(1)

    if args.level2tef:
        no_variables=False
        clb=int(args.level2tef)
        items=[]
        if clb!=None:
            print(colored(f'\nCLB {clb} to TEF:','green'))
            items.append(['Read','Write','Listen','Speak'])
            tef=clb2tef(clb)
            items.append([str(tef[0])+'-'+str(tef[1]),str(tef[2])+'-'+str(tef[3]),str(tef[4])+'-'+str(tef[5]),str(tef[6])+'-'+str(tef[7])])
            utils.printFList2D(items)
        else:
            print('You should input clb level...')
            exit(1)        

    if args.level2tcf:
        no_variables=False
        clb=int(args.level2tcf)
        items=[]
        if clb!=None:
            print(colored(f'\nCLB {clb} to TCF:','green'))
            items.append(['Read','Write','Listen','Speak'])
            tcf=clb2tef(clb)
            items.append([str(tcf[0])+'-'+str(tcf[1]),str(tcf[2])+'-'+str(tcf[3]),str(tcf[4])+'-'+str(tcf[5]),str(tcf[6])+'-'+str(tcf[7])])

            utils.printFList2D(items)
        else:
            print('You should input clb level...')
            exit(1)

    if args.ielts2level:
        no_variables=False
        r=float(args.ielts2level[0])
        w=float(args.ielts2level[1])
        l=float(args.ielts2level[2])
        s=float(args.ielts2level[3])
        items=[]
        clb, [clb_r, clb_w, clb_l, clb_s] = getCLBs([r,w,l,s],'IELTS')
        items.append(['IELTS','Read','Write','Listen','Speak'])
        items.append(['',r,w,l,s])
        items.append(['CLB','Read','Write','Listen','Speak'])
        items.append([clb,clb_r,clb_w,clb_l,clb_l])
        utils.printFList2D(items)

    if args.celpip2level:
        no_variables=False
        r=int(args.celpip2level[0])
        w=int(args.celpip2level[1])
        l=int(args.celpip2level[2])
        s=int(args.celpip2level[3])
        clb,[clb_r, clb_w, clb_l, clb_s] = getCLBs([r,w,l,s],"CELPIP")
        items=[]
        items.append(['CELPIP','Read','Write','Listen','Speak'])
        items.append(['',r,w,l,s])
        items.append(['CLB','Read','Write','Listen','Speak'])
        items.append([clb,clb_r,clb_w,clb_l,clb_l])
        utils.printFList2D(items)
    
    if args.tef2level:
        no_variables=False
        r=int(args.tef2level[0])
        w=int(args.tef2level[1])
        l=int(args.tef2level[2])
        s=int(args.tef2level[3])
        clb, [clb_r, clb_w, clb_l, clb_s] = getCLBs([r,w,l,s],'TEF')
        items=[]
        items.append(['TEF','Read','Write','Listen','Speak'])
        items.append(['',r,w,l,s])
        items.append(['CLB','Read','Write','Listen','Speak'])
        items.append([clb,clb_r,clb_w,clb_l,clb_s])
        utils.printFList2D(items)
    
    if args.tcf2level:
        no_variables=False
        r=int(args.tcf2level[0])
        w=int(args.tcf2level[1])
        l=int(args.tcf2level[2])
        s=int(args.tcf2level[3])
        clb, [clb_r, clb_w, clb_l, clb_s] = getCLBs([r,w,l,s],'TCF')
        items=[]
        items.append(['TCF','Read','Write','Listen','Speak'])
        items.append(['',r,w,l,s])
        items.append(['CLB','Read','Write','Listen','Speak'])
        items.append([clb,clb_r,clb_w,clb_l,clb_s])
        utils.printFList2D(items)
    
    if args.level:
        no_variables=False
        clb=int(args.level)
        items=[['Test','Read','Write','Listen','Speak']]
        if clb!=None:
            ielts=clb2ielts(clb)
            ielts.insert(0,"IELTS")
            tef=clb2tef(clb)
            tef.insert(0,'TEF')
            tcf=clb2tcf(clb)
            tcf.insert(0,'TCF')
            items.append(ielts)
            items.append(['CELPIP',clb,clb,clb,clb])
            items.append([tef[0],str(tef[1])+'-'+str(tef[2]),str(tef[3])+'-'+str(tef[4]),str(tef[5])+'-'+str(tef[6]),str(tef[7])+'-'+str(tef[8])])
            items.append([tcf[0],str(tcf[1])+'-'+str(tcf[2]),str(tcf[3])+'-'+str(tcf[4]),str(tcf[5])+'-'+str(tcf[6]),str(tcf[7])+'-'+str(tcf[8])])
            print(colored(f"\nCLB level {clb} equivalents to:","green"))
            utils.printFList2D(items)
        else:
            print('You should input clb level...')
            exit(1)
   
    if no_variables:
        print('CLB->IELTS')
        clb_ielts_table.insert(0,['Level','Read','Write','Listen','Speak'])
        utils.printFList2D(clb_ielts_table)
        
        print('CLB->TEF')
        tef=[['Level','Read','Write','Listen','Speak']]
        for t in tef_clb_table:
            tef.append([t[0],str(t[1])+'-'+str(t[2]),str(t[3])+'-'+str(t[4]),str(t[5])+'-'+str(t[6]),str(t[7])+'-'+str(t[8])])
        utils.printFList2D(tef)
        
        print('CLB->TCF')
        tef=[['Level','Read','Write','Listen','Speak']]
        for t in tcf_clb_table:
            tef.append([t[0],str(t[1])+'-'+str(t[2]),str(t[3])+'-'+str(t[4]),str(t[5])+'-'+str(t[6]),str(t[7])+'-'+str(t[8])])
        utils.printFList2D(tef)

if __name__=='__main__':
    main()



#!/Users/jacky/tools/venv/bin/python3.9

# This app is developed for creating a new client/project with necessary structure, folders, and files. 

import os,argparse,json,shutil
from datetime import date
from config.system import ROOT_DIR

def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)

def main():
    parser=argparse.ArgumentParser(description="For createing client/project with necessary structure, folders, and files")
    
    parser.add_argument("-cc", "--client", help="create a client working directory")
    parser.add_argument("-cp", "--project", help="create a project working directory ")

    args = parser.parse_args()

    if args.client:
        clientName=args.client
        if os.path.isdir(clientName):
            print(f'{clientName} is existing... ')
            return 
        else:
            os.makedirs(clientName)

        source=os.path.join(ROOT_DIR,"templates/client")
        copytree(source,clientName,symlinks=False,ignore=None)
        print(f'A working directory for client {clientName} is ready...')
        return 
    
    if args.project:
        # make project name, exp: 20211001-lmia
        today=date.today()
        project=(str)(today.year)+str(today.month).zfill(2)+str(today.day).zfill(2)+"-"+args.project

        # create project folder
        if os.path.isdir(project):
            print(f'The project {project} is existing...')
            return 
        else:
            os.makedirs(project)
        
        # get template folder
        jsonDir=os.path.join(ROOT_DIR,"templates/templates.json")
        with open(jsonDir) as jf:
            dirs=json.load(jf)
        source=os.path.join(ROOT_DIR,dirs[args.project])

        # copy template contents to working project
        copytree(source,project,symlinks=False,ignore=None)
        print(f'A working directory for project {project} is ready...')
        return 

if __name__=="__main__":
    main()




from ast import Str
from pydantic import BaseModel,validator,EmailStr
from datetime import date
from typing import Optional,List,Union
from model.common.employerbase import EmployerBase
from model.common.jobofferbase import JobofferBase
from model.common.contact import ContactBase
from pydantic.class_validators import root_validator

from model.common.utils import Duration


# employer classes
class General(EmployerBase):
    company_intro:str
    business_intro:str
    recruit_email:EmailStr

class Contact(ContactBase):
    position:Optional[str]
    
#personal classes
class JobOffer(JobofferBase):
    offer_date:date

    @property
    def date_of_offer(self):
        return self.offer_date.strftime("%b %d, %Y")
    
    @property
    def requirements(self):
        return [r for r in [self.specific_edu_requirement,self.skill_experience_requirement,*self.other_requirements] if r is not None]

class PersonalAssess(BaseModel):
    work_experience_brief:str
    education_brief:str
    competency_brief:str
    language_brief:Optional[str]
    performance_remark:Optional[str]
    
    @property
    def why_qualified(self):
        qualifications=[self.work_experience_brief,self.education_brief,self.competency_brief,self.language_brief,self.performance_remark]
        return [q for q in qualifications if q is not None]

# lmia classes    
class LmiaCase(BaseModel):
    purpose_of_lmia:str
    stream_of_lmia:str
    number_of_tfw:int
    area_median_wage:float
    provincial_median_wage:float
    duration_number:int
    duration_unit:str
    duration_reason:str
    has_attestation:bool
    
    @property
    def purpose_say(self):
        return self.purpose_of_lmia
    
    @property
    def stream_say(self):
        return self.stream_of_lmia
    
    @property
    def duration_say(self):
        return str(self.duration_number)+" "+self.duration_unit+"s" if self.duration_number>1 else str(self.duration_number)+" "+self.duration_unit
    
    @property
    def stars(self):
        return 2 if self.stream_of_lmia.lower() in ['low wage stream','lws'] else 4

class Finance(BaseModel):
    year:int
    revenue:float
    net_income:float
    retained_earning:float
    
    @property
    def formatted_revenue(self):
        return '{:,.0f}'.format(self.revenue)
    
    @property
    def formatted_net_income(self):
        return '{:,.0f}'.format(self.net_income)
    
    @property
    def formatted_retained_earning(self):
        return '{:,.0f}'.format(self.retained_earning)
    
        
class Lmi(BaseModel):
    brief_benefit:str
    job_creation_benefit:Optional[str]
    skill_transfer_benefit:Optional[str]
    fill_shortage_benefit:Optional[str]
    other_benefit:Optional[str]

    @property
    def benefits(self):
        return [b for b in [self.job_creation_benefit,self.skill_transfer_benefit,self.fill_shortage_benefit,self.other_benefit] if b is not None]

# RCIC
class Rcic(BaseModel):
    first_name:str
    last_name:str
    sex:str
    rcic_number:str
    company:str
    
    @property
    def name(self):
        return self.first_name+' '+self.last_name
        
    
class Emp5593(BaseModel):
    pass

class Emp5626(BaseModel):
    is_seasonal:bool
    start_month:Optional[str]
    end_month:Optional[str]
    last_canadian_number:Optional[int]
    last_tfw_number:Optional[int]
    current_canadian_number:int
    current_tfw_number:int
    tp_waivable:bool
    waive_creteria:Optional[str]
    has_finished_tp:bool
    finished_tp_result:Optional[str]
    activity1_title:Optional[str]
    activity1_decription:Optional[str]
    activity1_outcome:Optional[str]
    activity1_commnet:Optional[str]
    activity2_title:Optional[str]
    activity2_decription:Optional[str]
    activity2_outcome:Optional[str]
    activity2_commnet:Optional[str]
    activity3_title:Optional[str]
    activity3_decription:Optional[str]
    activity3_outcome:Optional[str]
    activity3_commnet:Optional[str]
    activity4_title:Optional[str]
    activity4_decription:Optional[str]
    activity4_outcome:Optional[str]
    activity4_commnet:Optional[str]
    activity5_title:Optional[str]
    activity5_decription:Optional[str]
    activity5_outcome:Optional[str]
    activity5_commnet:Optional[str]

class Emp5627(BaseModel):
    provide_accommodation:bool
    description:Optional[str]
    rent_unit:str
    rent_amount:float
    accommodation_type:str
    explain:Optional[str]
    bedrooms:int
    people:int
    bathrooms:int
    other:Optional[str]



from typing import Union
from webform.models.definition import Action
from webform.models.bcpnpmodel_reg import BcpnpModelReg, BcpnpEEModelReg
from webform.bcpnp.dashboard import DashboardReg
from model.common.employmenthistory import EmploymentHistory
from datetime import datetime


class Employment:
    def __init__(self, person: Union[BcpnpModelReg, BcpnpEEModelReg]):
        self.person = person

    def has_work_experience(self, jobs):
        return [
            {
                "action_type": Action.Radio.value,
                "label": "Do you have directly related work experience at the same or greater NOC level in the past 10 years, including the Job Offer?",
                "id": "#BCPNP_App_Job_Work_ExperienceRelated-Yes"
                if len(jobs) > 0
                else "#BCPNP_App_Job_Work_ExperienceRelated-No",
            }
        ]

    def block(self, job, index):
        id_present = (
            "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(3) > div > uf-panel > div > uf-panel-body > div > div > div.ng-pristine.ng-untouched.ng-valid.ng-isolate-scope.ng-not-empty > div > uf-row:nth-child(4) > div > uf-date:nth-child(2) > div > div.form-inline.present-checkbox-container > div > label"
            if index == 0
            else f"body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(3) > div > uf-panel > div > uf-panel-body > div > div > div.ng-pristine.ng-untouched.ng-valid.ng-isolate-scope.ng-not-empty > div:nth-child({index+1}) > uf-row:nth-child(4) > div > uf-date:nth-child(2) > div > div.form-inline.present-checkbox-container > div"
        )
        return [
            {
                "action_type": Action.Input.value,
                "label": "Job title",
                "id": f"#BCPNP_App_Work_Title-{index}",
                "value": job.job_title,
                "required": True,
                "length": 200,
            },
            {
                "action_type": Action.Input.value,
                "label": "NOC (4 digits)",
                "id": f"#BCPNP_App_Work_NOC-{index}",
                "value": job.noc_code,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Radio.value,
                "label": "Job hours ",
                "id": f"#BCPNP_App_Work_JobHours-{index}-full-time"
                if job.weekly_hours >= 30
                else f"#BCPNP_App_Work_JobHours-{index}-part-time",
            },
            {
                "action_type": Action.Input.value,
                "label": "Start date",
                "id": f"#BCPNP_App_Work_From-{index}",
                "value": job.start_date.strftime("%Y-%m-%d"),
                "required": True,
                "length": 10,
            },
            {
                "action_type": Action.Input.value,
                "label": "End date",
                "id": f"#BCPNP_App_Work_To-{index}",
                "value": datetime.today().strftime("%Y-%m-%d")
                if job.is_present
                else job.end_date.strftime("%Y-%m-%d"),
                "required": True,
                "length": 10,
            },
            {
                "action_type": Action.Checkbox.value,
                "label": "Present",
                "id": id_present,
                "value": True if job.end_date else False,
            },
            {
                "action_type": Action.Input.value,
                "label": "Company name",
                "id": f"#BCPNP_App_Work_Company-{index}",
                "value": job.company,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Radio.value,
                "label": "Was this work in Canada? ",
                "id": f"#BCPNP_App_WorkExp_InCan-{index}-Yes"
                if job.country == "Canada"
                else f"#BCPNP_App_WorkExp_InCan-{index}-No",
            },
        ]

    def repeat_blocks(self, jobs):
        blocks = []
        for index, job in enumerate(jobs):
            blocks.append(self.block(job, index))
        return blocks

    def fill(self):
        # get highest edu level
        jobs = EmploymentHistory(self.person.employment).qualified_employment(
            "bcpnp_qualified"
        )
        employment = [
            {
                "action_type": Action.RepeatSection.value,
                "button_id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(3) > div > uf-panel > div > uf-panel-body > div > div > div:nth-child(2) > uf-clone-repeatable > a > i",
                "value": self.repeat_blocks(jobs),
            }
        ]
        # dashboard = DashboardReg()
        actions = (
            # dashboard.jump("Work Experience")
            self.has_work_experience(jobs)
            + employment
            # + dashboard.save
        )
        return [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Work Experience",
                "actions": actions,
                "id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > ul > li:nth-child(4) > a",
                "next_page_tag":"#syncA_App_Emp_Comp_LegalName"
            }
        ]



from typing import List, Optional
from model.common.address import Address
from model.common.phone import Phone
from model.common.trperson import (
    COR,
    PersonId,
    Personal,
    Marriage,
    Education,
    Employment,
    Family,
)
from model.common.tr import TrCaseIn, TrBackground, WpInCanada
from model.common.commonmodel import CommonModel


class M5710Model(CommonModel):
    personal: Personal
    marriage: Marriage
    personid: List[PersonId]
    address: List[Address]
    education: List[Education]
    employment: List[Employment]
    family: List[Family]
    phone: List[Phone]
    cor: List[COR]
    trcasein: TrCaseIn
    wpincanada: WpInCanada
    trbackground: TrBackground

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(["excel/tr.xlsx", "excel/pa.xlsx"])
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())



import os,sys
# Get project's home directory, 
BASEDIR=os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
# All data directory
DATADIR=os.path.abspath(os.path.join(BASEDIR,'data'))
# Insert the BASEDIR to system path
sys.path.insert(0,BASEDIR)




from .context import DATADIR
from email.policy import default
from textwrap import indent
from typing import List,Optional
from model.experience.resumedata import Personal,Language,PersonalAssess,Education, Employment
from model.common.commonmodel import CommonModel
from model.common.phone import Phone,Phones 
from model.common.address import Address,Addresses
from datetime import date
from model.common.wordmaker import WordMaker
import os,json

class ResumeModel(CommonModel):
    personal:Personal
    phone:List[Phone]
    personalassess:PersonalAssess
    education:List[Education]
    language:Optional[List[Language]]
    employment:Optional[List[Employment]]
    address:List[Address]
    
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            excels=self.getExcels(['excel/pa.xlsx'])
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())
        
class ResumeModelDocxAdapater():
    """This is an adapater to bridging resume model data and docx data
    """

    def __init__(self,resume_obj: ResumeModel):
        # get original resume obj, which will be used to generate some value based on it's object methods. 
        self.resume_obj=resume_obj
        #从列表中挑选element出来，并替换该列表
        phones=Phones(self.resume_obj.phone)
        addresses=Addresses(self.resume_obj.address)
        
        self.resume_obj.phone=phones.PreferredPhone
        self.resume_obj.address=addresses.PreferredAddress

    def make(self,output_docx,template_no=None):
        template_no = template_no or 1
        filename="word/resume-regular"+str(template_no)+".docx"
        template_path=os.path.abspath(os.path.join(DATADIR,filename))    
        wm=WordMaker(template_path,self.resume_obj,output_docx)
        wm.make()






# from context import BASEDIR
from source.excel import Excel
from bin.common import ExcelMaker,valid_programs
import argparse
from functools import reduce
from termcolor import colored
import json

def main():
    parser=argparse.ArgumentParser(description="For munipulate excel and get data source...")
    
    parser.add_argument("-e", "--excel", help="Input excel name")
    parser.add_argument("-a", "--add", help="add up excels(can be more than 2), and output to another excel",nargs="+")
    parser.add_argument("-s", "--sub", help="sub excels(can be more than 2), and output to another excel",nargs="+")
    parser.add_argument("-c", "--copy", help="copy later excels' common contents to first excel(can be more than 2), and output to another excel",nargs="+")
    parser.add_argument("-t", "--to", help="Output file name")
    parser.add_argument("-po", "--protection_off", help="Protection off. default is On",action='store_true')
    parser.add_argument("-j", "--json", help="need json output",action='store_true')
    parser.add_argument("-d", "--dict", help="need dict output",action='store_true')
    parser.add_argument("-p", "--program", help="input program code. 5257 for trv, 1294 or 1295 for sp or wp outside of Canada. 5708/5709/5710 for vr/sp/wp in Canada ")
    
    
    args = parser.parse_args()
    protection=False if args.protection_off else True

    if args.add and args.to:
        if len(args.add)<2:
            raise ValueError("Must more than two excel files after -a ")
        try:
            excel_objs=list(map(lambda x:Excel(x), args.add))
            e=reduce(lambda a,b:a+b, excel_objs)
            e.makeExcel(args.to,protection=protection)
            print(colored(f'{args.to} saved',"green"))
            if args.json:
                print(e.json)
            if args.dict:
                print(e.dict)  
        except Exception as e:
            print(colored(f'{str(e)}','red'))
            
    if args.sub and args.to:
        if len(args.sub)<2:
            raise ValueError("Must more than two excel files after -a ")
        try:
            excel_objs=list(map(lambda x:Excel(x), args.sub))
            e=reduce(lambda a,b:a-b, excel_objs)
            e.makeExcel(args.to,protection=protection)
            print(colored(f'{args.to} saved',"green"))
            if args.json:
                print(e.json)
            if args.dict:
                print(e.dict) 
            
        except Exception as e:
            print(colored(f'{str(e)}','red'))
    
    if args.copy and args.to:
        if len(args.copy)<2:
            raise ValueError("Must  more than two excel files after -c ")
        try:
            excel_objs=list(map(lambda x:Excel(x), args.copy))
            e=reduce(lambda a,b:a.copy(b), excel_objs)
            e.makeExcel(args.to,protection=protection)
            print(colored(f'{args.to} saved',"green"))
            if args.json:
                print(e.json)
            if args.dict:
                print(e.dict) 
            
        except Exception as e:
            print(colored(f'{str(e)}','red'))
        
            
    if args.program and args.to:
        if args.program not in valid_programs:
            print(f'Your input program code {args.program} is not valid. ')
            return
        excel_file=ExcelMaker(args.program)
        excel_file.makeExcelBasedOnModel(args.to)
    
    if args.excel and args.to and args.json:
        excel=Excel(args.excel)
        with open(args.to,'w') as fp:
            json.dump(excel.dict,fp,indent=3,default=str)
        print(colored(f"{args.to} is saved","green"))
        
if __name__=="__main__":
    main()






# from context import BASEDIR
import argparse
from tabulate import tabulate
from termcolor import colored
import os
import json
import requests
from dotenv import dotenv_values
from assess.engine.quickassessmodel import QuickAssessModel
from datetime import date,datetime
import warnings
warnings.filterwarnings("ignore")

api_url = "https://jackyzhang.pro/assess/quick/"

def education_level(level):
    edu_level={
        'Doctor':["Doctor"],
        'Master':["Master"],
        'Bachelor': ["Post-graduate diploma",'Bachelor'],
        'Diploma':["Associate","Diploma/Certificate"],
        'Secondary':["High School"],
        'Other':["Less than high school"]
        }
    for el,l in edu_level.items():
        if level in l:
            return el;


def getData(excel_file):
    data = QuickAssessModel(
        excels=[excel_file]).dict()
    data = data['facts']
    level=data['level']

    return_data = {
        "basic": {
            "dob": data['dob'].strftime("%Y-%m-%d"),
            "email": "jacky@gmail.com",
            "phone": "7783215110",
            "lastName": "Zhang",
            "firstName": "Jacky",
            "birthCity": "Nanjing",
            "birthCountry": "China"
        },
        "language": {
            "date": date.today().strftime('%Y-%m-%d'),
            "reading": data['reading'] if data['reading'] else 0,
            "writting": data['writting'] if data['writting'] else 0,
            "speaking": data['speaking'] if data['speaking'] else 0,
            "listening": data['listening'] if data['listening'] else 0,
            "testFormat": data['test_format'] if data['test_format'] else "IELTS"
        },
        "education": {
            "level": education_level(level),
            "country": data['edu_country'],
            "province": data['edu_province'],
            "institute": "NJU",
            "graduateDate": data['graduate_date'].strftime('%Y-%m-%d')
        },
        "family": [
            {
                "dob": data['relative_dob'].strftime("%Y-%m-%d") if data['relative_dob'] else datetime.today().strftime("%Y-%m-%d"),
                "city": "Vancouver",
                "name": "Li",
                "province": data['relative_province']if data['relative_province'] else "BC",
                "canadaStatus": data['canada_status'] if data['canada_status'] else "PR",
                "relationship": data['relationship'] if data['relationship'] else "spouse"
            }
        ],
        "assets": {
            "liquidAssets": data['liquid_assets'] if data['liquid_assets'] else 0,
            "netWorth": data['net_worth'] if data['net_worth'] else 0
        },
        "workExperience": {
            "startDate": data['start_date'].strftime('%Y-%m-%d'),
            "endDate": data['end_date'].strftime('%Y-%m-%d'),
            "noc": data['work_noc'],
            "term": data['term'],
            "country": data['work_country'],
            "jobTitle": data['job_title'],
            "province": data['work_province'],
            "sharePercentage": data['share_percentage']
        },
        "adaptation": {
            "noc": data['joboffer_noc'] if data['joboffer_noc'] else '1123',
            "title": "Marketing specialist",
            "jobOffer": data['job_offer'],
            "province": data['joboffer_province'] if data['joboffer_province'] else "BC"
        }
    }
    return return_data


def getSolutions(data):
    path=os.path.abspath(os.path.join(os.path.dirname(__file__), ".env"))
    config = dotenv_values(path)
    token = config['TOKEN']
    headers = {
        'Authorization': f"Token {token}",
        'Content-Type': 'application/json'
    }
    response = requests.post(api_url, json=data, headers=headers)
    if response.status_code == 200:
        return response.json()
    else:
        print(
            f' Response status code is {response.status_code}, and there is no feedback')


def showData(solutions,args):
    if solutions:
        print(colored('You may be qualified for the Canadian immigration programs as the following\n', 'green'))
        for prog in solutions:
            print(colored("* {program} *".format(**prog), 'green'))
            print(colored("Stream: ","green"),colored("{stream}".format(**prog), 'white'))
            if args.description: print(colored("Description: ","green"),colored("{description}".format(**prog), 'white'))
            if args.remark: print(colored("Remark: ","green"),colored("{remark}".format(**prog), 'white'))
    else:
        print(colored("You are not qualified for any program","red"))


def main():
    parser = argparse.ArgumentParser(
        description="used for processing everything noc related")
    parser.add_argument("-e", "--excel", help="input excel file")
    parser.add_argument(
        "-j", "--json", help="print json data ", action='store_true')
    parser.add_argument(
        "-d", "--description", help="print program description data ", action='store_true')
    parser.add_argument(
        "-r", "--remark", help="print program remark data ", action='store_true')

    args = parser.parse_args()

    if args.excel:
        data = getData(args.excel)
        solutions=getSolutions(data)
        showData(solutions,args)
        if args.json:
            print(json.dumps(data, indent=3))


if __name__ == "__main__":
    main()



from logging import root
import xml.etree.ElementTree as ET
import xmltodict
import copy

#几个对应点
# 1. data source的原始值key:value。该原始字典的keys为核心变量列表，该变量在get和set两个部分是核心变量。在获取xml信息中，提供keys列表去获取数据。 在get中，提供字典去设置值。
# 2. 核心变量对应的在字典中的位置。 key可以由核心变量+‘_path',赋值为一个list，里面包含从字典从外到内的keys
# 3. 核心变量对应的的data source的keys。 可以由核心变量+’_pairs‘,赋值为对应的excel的keys。  

class pdfformmodel(object):
    
    def __init__(self) -> None:
        self.xml_data={}
    
    def _getDict(self,data_dict,path):
        new_dict=data_dict
        for p in path:
            new_dict=new_dict[p]
        return new_dict
    
    #get info from xml, and not in excluded keys
    def _getItems(self,path,excluded_keys=None,pairs=None):
        rows=self._getDict(self.data_dict,path)
        temp_list=[]
        for row in rows:
            keys=[key for key in row.keys() if excluded_keys and key not in excluded_keys]
            # convert keys based on pairs
            temp_list.append({pairs.get(k):v for k,v in row.items() if pairs and k in keys})
        return temp_list
    
    # convert xml to dict
    def getXmlDict(self,source_xml):
        tree=ET.parse(source_xml)
        root=tree.getroot()
        xmlstr = ET.tostring(root, encoding='utf8', method='xml')
        data_dict = dict(xmltodict.parse(xmlstr))
        return data_dict
    
    # get xml info in excel source data format 
    def getXmlData(self,source_xml):
        self.data_dict=self.getXmlDict(source_xml)    
        
        #子类必须定义和paths的keys相对应的pairs，excludes keys。 此处应该根据paths的key，自动生成对应的变量名，并自动查找其需要的值
        for variable, path in self.paths.items(): #variable 是大类， path是其对应在form中keys chain的列表
            # variable_rows=self._getDict(self.data_dict,path)
            excluded_keys=self.excluded_keys.get(variable)
            pair=self.pairs.get(variable)
            self.xml_data[variable]=self._getItems(path,excluded_keys=excluded_keys,pairs=pair)

        return self.xml_data
    
    # fill xml template file with xml_data 
    def makeXml(self,template,output,context=None):
        context=context if context else self.xml_data
        # if len(self.xml_data)==0:
        #     raise ValueError('No data. Please getXmlData first...')
        o=[]
        with open(template) as x:
            s=x.read()
        out=''.join(s.format(**context))
        with open(output,'w')as w:
            w.write(out)
            print(f' {output} Saved')








from pydantic import BaseModel,EmailStr,validator
from typing import Optional
from model.common.utils import normalize

    
class EmployerBase(BaseModel):
    legal_name:str
    operating_name:Optional[str]
    website:Optional[str]
    
    _normalize_legal_name=validator('legal_name',allow_reuse=True,check_fields=False)(normalize)
    _normalize_operating_name=validator('operating_name',allow_reuse=True,check_fields=False)(normalize)
    
    @property
    def preferred_name(self):
        return self.operating_name if self.operating_name else self.legal_name



def main():
    p1={
        'legal_name':'jacky',
        'operating_name':"zhang",
        'website':"China"
    }
    jacky=EmployerBase(**p1)
    print(jacky.preferred_name)

if __name__=="__main__":
    main()
    


from turtle import position
from typing import List
from model.common.commonmodel import CommonModel
from model.bcpnp.data import Contact, General
from model.common.person import Person
from model.common.contact import Contacts
from datetime import date
import os
from model.common.xmlfiller import XmlFiller


class Personal(Person):
    def __str__(self):
        return self.full_name
        
class RepFormModel(CommonModel):
    general:General
    contact:List[Contact]
    personal:Personal

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            excels=self.getExcels(['excel/er.xlsx','excel/pa.xlsx'])
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())
    
    @property
    def selected_contact(self):
        contacts=Contacts(self.contact)
        return contacts.preferredContact
    
    @property
    def person(self):
        return {
            "first_name":self.personal.first_name,
            "last_name":self.personal.last_name,
            "dob":self.personal.dob.strftime('%d-%b-%Y'),
        }

class RepFormDocxAdaptor():
    def __init__(self,repform_obj:RepFormModel):
        self.repform_obj=repform_obj
        
    def re_generate_dict(self):
        summary_info={
            'personal':self.repform_obj.person,
            'date_signed':date.today(),
            "contact_last_name":self.repform_obj.selected_contact.last_name,
            "contact_first_name":self.repform_obj.selected_contact.first_name,
        }
        return {**self.repform_obj.dict(),**summary_info}

    def make(self,output_xml):
        path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path=path+"/template/xml/bcpnp_rep.xml"            
        xf=XmlFiller(template_path,self.re_generate_dict())
        xf.save(output_xml)


from webform.models.definition import Action
from webform.models.bcpnpmodel_app import BcpnpModelApp, BcpnpEEModelApp
from webform.bcpnp.dashboard import DashboardApp
from typing import Union
from webform.bcpnp.data import city_map
from webform.bcpnp.data import status_in_canada_map, country_map, workpermit_type_map


class Applicant:
    def __init__(self, person: Union[BcpnpModelApp, BcpnpEEModelApp]):
        self.person = person

    def fill(self):
        intended_residence_place = [
            {
                "action_type": Action.Select.value,
                "label": "In which city/town will you live after nomination?",
                "id": "#syncA_App_IntendedResidence",
                "value": city_map[self.person.bcpnp.intended_city],
            }
        ]

        # Current/previous applications
        q1 = [
            {
                "action_type": Action.Radio.value,
                "label": "Do you have another active application or registration with BC PNP?",
                "id": "#BCPNP_App_ActiveApplication-Yes"
                if self.person.bcpnp.q1
                else "#BCPNP_App_ActiveApplication-No",
            }
        ]
        q1_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq1",
                    "value": self.person.bcpnp.q1_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q1
            else []
        )

        q2 = [
            {
                "action_type": Action.Radio.value,
                "label": "Have you had another Skills Immigration or Entrepreneur Immigration registration or application refused or approved with BC PNP?",
                "id": "#BCPNP_App_RejectedPrevPNP-Yes"
                if self.person.bcpnp.q2
                else "#BCPNP_App_RejectedPrevPNP-No",
            }
        ]
        q2_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq2",
                    "value": self.person.bcpnp.q2_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q2
            else []
        )

        q3 = [
            {
                "action_type": Action.Radio.value,
                "label": "Do you have an active application for provincial nomination or permanent residence under any other federal or provincial program?",
                "id": "#BCPNP_App_ActivePNPAppReg-Yes"
                if self.person.bcpnp.q3
                else "#BCPNP_App_ActivePNPAppReg-No",
            }
        ]
        q3_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq3",
                    "value": self.person.bcpnp.q3_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q3
            else []
        )

        q4 = [
            {
                "action_type": Action.Radio.value,
                "label": "Have you ever had an application rejected for provincial nomination or permanent residence in Canada under any other federal or provincial program? ",
                "id": "#BCPNP_App_RejectedApplication-Yes"
                if self.person.bcpnp.q4
                else "#BCPNP_App_RejectedApplication-No",
            }
        ]
        q4_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq4",
                    "value": self.person.bcpnp.q4_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q4
            else []
        )

        q5 = [
            {
                "action_type": Action.Radio.value,
                "label": "Have you ever had an application rejected for a Canadian visitor visa, study permit, or temporary work permit?",
                "id": "#BCPNP_App_RejectedVisaPermit-Yes"
                if self.person.bcpnp.q5
                else "#BCPNP_App_RejectedVisaPermit-No",
            }
        ]
        q5_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq5",
                    "value": self.person.bcpnp.q5_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q5
            else []
        )

        q6 = [
            {
                "action_type": Action.Radio.value,
                "label": "Have you made a claim for refugee protection in Canada, or have you been refused refugee status in Canada? ",
                "id": "#BCPNP_App_RefusedRefugee-Yes"
                if self.person.bcpnp.q6
                else "#BCPNP_App_RefusedRefugee-No",
            }
        ]
        q6_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq6",
                    "value": self.person.bcpnp.q6_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q6
            else []
        )

        q7 = [
            {
                "action_type": Action.Radio.value,
                "label": "Are you under a removal order from Canada (e.g. departure order or exclusion order)?",
                "id": "#BCPNP_App_RemovalOrder-Yes"
                if self.person.bcpnp.q7
                else "#BCPNP_App_RemovalOrder-No",
            }
        ]
        q7_details = (
            [
                {
                    "action_type": Action.Input.value,
                    "label": "Provide details",
                    "id": "#BCPNP_App_CurPrevApplicationsDetailsq7",
                    "value": self.person.bcpnp.q7_explaination,
                    "required": True,
                    "length": 200,
                }
            ]
            if self.person.bcpnp.q7
            else []
        )

        # Current status
        current_in_canada = [
            {
                "action_type": Action.Radio.value,
                "label": "Are you currently in Canada?",
                "id": "#syncA_App_InCanada-Yes"
                if self.person.status.current_country == "Canada"
                else "#syncA_App_InCanada-No",
            }
        ]
        status_or_country = (
            [
                {
                    "action_type": Action.Select.value,
                    "label": "Indicate your current status in Canada",
                    "id": "#syncA_App_InCanada_Status",
                    "value": status_in_canada_map[
                        self.person.status.current_country_status
                    ],
                }
            ]  # BCPNP_App_CurResidence_Country
            if self.person.status.current_country == "Canada"
            else [
                {
                    "action_type": Action.Select.value,
                    "label": "Indicate your current country of residence",
                    "id": "#BCPNP_App_CurResidence_Country",
                    "value": country_map[self.person.status.current_country],
                }
            ]
        )

        status_details = []

        def uci(id):
            return {
                "action_type": Action.Input.value,
                "label": "IRCC Client ID/UCI",
                "id": id,
                "value": self.person.personal.uci,
                "required": True,
                "length": 20,
            }

        def start_date(id):
            return {
                "action_type": Action.Input.value,
                "label": "Date of issue",
                "id": id,
                "value": self.person.status.current_status_start_date,
                "required": True,
                "length": 20,
            }

        def end_date(id):
            return {
                "action_type": Action.Input.value,
                "label": "Valid until",
                "id": id,
                "value": self.person.status.current_status_end_date,
                "required": True,
                "length": 20,
            }

        match self.person.status.current_country_status:
            case "Student":
                status_details = [
                    uci("#syncA_App_StudyPermit_ClientID"),
                    start_date("#BCPNP_App_StudyPermit_DateSigned"),
                    end_date("#syncA_App_StudyPermit_ValidUntil"),
                ]
            case "Worker":
                workpermit_type = {
                    "action_type": Action.Select.value,
                    "label": "Your work permit type",
                    "id": "#BCPNP_App_WorkPermit_Info",
                    "value": workpermit_type_map[
                        self.person.status.current_workpermit_type
                    ],
                }
                status_details = [
                    workpermit_type,
                    uci("#syncA_App_WorkPermit_ClientID"),
                    start_date("#BCPNP_App_WorkPermit_DateSigned"),
                    end_date("#syncA_App_WorkPermit_ValidUntil"),
                ]
            case "Visitor":
                has_visitor = {
                    "action_type": Action.Radio.value,
                    "label": "Do you have a visitor record? ",
                    "id": "#BCPNP_App_VisitorRecord-Yes"
                    if self.person.status.has_vr
                    else "#BCPNP_App_VisitorRecord-No",
                }
                enter_date = {
                    "action_type": Action.Input.value,
                    "label": "Date entered Canada",
                    "id": "#BCPNP_App_Visitor_EnteredCanada",
                    "value": self.person.status.last_entry_date,
                    "required": True,
                    "length": 20,
                }
                status_details = [
                    has_visitor,
                    uci("#BCPNP_App_Visitor_ClientID"),
                    enter_date,
                    start_date("#BCPNP_App_Visitor_DateSigned"),
                    end_date("#BCPNP_App_Visitor_ValidUntil"),
                ]
            case "Other":
                status_details = [
                    {
                        "action_type": Action.Input.value,
                        "label": "If other, please specify:",
                        "id": "#BCPNP_App_InCanada_Status_Other",
                        "value": self.person.status.other_status_explaination,
                        "required": True,
                        "length": 200,
                    }
                ]

        # dashboard = DashboardApp()

        actions = (
            # dashboard.jump("Applicant")
            intended_residence_place
            + q1
            + q1_details
            + q2
            + q2_details
            + q3
            + q3_details
            + q4
            + q4_details
            + q5
            + q5_details
            + q6
            + q6_details
            + q7
            + q7_details
            + current_in_canada
            + status_or_country
            + status_details
            # + dashboard.save
        )
        return [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Applicant",
                "actions": actions,
                "id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > ul > li:nth-child(2) > a",
                "next_page_tag":"#BCPNP_App_EduNCAN-No"
            }
        ]



from webform.models.definition import Action
from webform.models.bcpnpmodel_reg import BcpnpModelReg, BcpnpEEModelReg
from webform.models.bcpnpmodel_app import BcpnpEEModelApp, BcpnpModelApp
from typing import Union
import base64

# This step includs actions of signing in, picking Skill Immigration automatically, select and confirm which stream to apply through
class Login:
    def __init__(
        self,
        person: Union[BcpnpModelReg, BcpnpEEModelReg, BcpnpEEModelApp, BcpnpModelApp],
    ):
        self.person = person

    def encode(self, password):
        password_bytes = password.encode("ascii")
        base64_bytes = base64.b64encode(password_bytes)
        return base64_bytes.decode("ascii")

    def login(self, initial=True):
        sign_in = [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Sign in",
                "actions": [
                    {
                        "action_type": Action.GotoPage.value,
                        "url": "https://www.pnpapplication.gov.bc.ca/user/sign-in",
                    },
                    {
                        "action_type": Action.Login.value,
                        "label": "Login",
                        "account": self.person.bcpnp.account or self.person.personal.user_id,
                        "password": self.encode(self.person.bcpnp.password) or self.encode(self.person.personal.password),
                        "account_element_id": "#userId",
                        "password_element_id": "#pass",
                        "login_button_id": "#form > div:nth-child(4) > div > input",
                    },
                ],
                "id": None,  #
            }
        ]

        pick_sk_immigration = [
            # pick skill immigration
            {
                "action_type": Action.WebPage.value,
                "page_name": "Pick skill immigration",
                "actions": [],
                "id": "body > div > main > div.layout-container > div > div > div > div:nth-child(1) > a",
                "next_page_tag": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(3) > ul > li:nth-child(5) > a"
            },
        ]

        # pick a stream of bcpnp
        case_stream = {
            "EE-Skilled Worker": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(1) > ul > li:nth-child(1) > a",
                "confirm": "#skills-express-skilled > form > div > label > input[type=checkbox]",
                "start": "#skills-express-skilled > form > input",
            },
            "EE-International Graduate": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(1) > ul > li:nth-child(2) > a",
                "confirm": "#skills-express-intl-grad > form > div > label > input[type=checkbox]",
                "start": "#skills-express-intl-grad > form > input",
            },
            "EE-International Post-Graduate": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(1) > ul > li:nth-child(3) > a",
                "confirm": "#skills-express-intl-postgrad > form > div > label > input[type=checkbox]",
                "start": "#skills-express-intl-postgrad > form > input",
            },
            "EE-Health Authority": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(1) > ul > li:nth-child(4) > a",
                "confirm": "#skills-express-health-care > form > div > label > input[type=checkbox]",
                "start": "#skills-express-health-care > form > input",
            },
            "Skilled Worker": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(3) > ul > li:nth-child(1) > a",
                "confirm": "#skills-skilled > form > div > label > input[type=checkbox]",
                "start": "#skills-skilled > form > input",
            },
            "International Graduate": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(3) > ul > li:nth-child(2) > a",
                "confirm": "#skills-intl-grad > form > div > label > input[type=checkbox]",
                "start": "#skills-intl-grad > form > input",
            },
            "Entry-Level and Semi-Skilled Worker": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(3) > ul > li:nth-child(3) > a",
                "confirm": "#skills-entry-level > form > div > label > input[type=checkbox]",
                "start": "#skills-entry-level > form > input",
            },
            "International Post-Graduate": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(3) > ul > li:nth-child(4) > a",
                "confirm": "#skills-intl-postgrad > form > div > label > input[type=checkbox]",
                "start": "#skills-intl-postgrad > form > input",
            },
            "Health Authority": {
                "link": "body > div > main > div.layout-container > div > div > div:nth-child(4) > div:nth-child(3) > ul > li:nth-child(5) > a",
                "confirm": "#skills-health-care > form > div > label > input[type=checkbox]",
                "start": "#skills-health-care > form > input",
            },
        }
        pick_stream = [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Pick skill immigration",
                "actions": [
                    {
                        "action_type": Action.Button.value,
                        "label": "Pick which stream to apply through",
                        "id": case_stream[self.person.bcpnp.case_stream]["link"],
                    },
                    {
                        "action_type": Action.Checkbox.value,
                        "label": "I confirm that I qualify all the category requirements",
                        "id": case_stream[self.person.bcpnp.case_stream]["confirm"],
                        "value":True
                    },
                ],
                "id": case_stream[self.person.bcpnp.case_stream]["start"],
                "next_page_tag": "body > div > main > div.layout-container > div > div > form > button"
            }
        ]

        confirm_profile = [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Confirm profile and representative",
                "actions": [],
                "id": "body > div > main > div.layout-container > div > div > form > button",
                "next_page_tag": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > ul > li.ng-scope.active > a"
            }
        ]

        continue_page = [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Continue my registration",
                "actions": [],
                "id": "body > div.page > main > div > div > div > div > div:nth-child(2) > div > div.col-sm-5.col-sm-offset-2.tile > div > div.buttons > a.btn.btn-primary",
                "next_page_tag":"body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > ul > li.ng-scope.active > a"
            },
        ]

        return (
            sign_in + pick_sk_immigration + pick_stream + confirm_profile
            if initial
            else sign_in + continue_page
        )



from datetime import date
from typing import Optional,List
from pydantic import BaseModel
from model.common.commonmodel import CommonModel
import os


class General(BaseModel):
    legal_name:str
    operating_name:Optional[str]
    website:Optional[str]
    establish_date:date
    num_pnps:int
    num_pnps_approved:int
    num_pnps_in_process:int
    has_lmia_approved:bool
    when_lmia_approved:Optional[date]
    last_revenue:float
    last_profit:float
    before_last_revenue:float
    before_last_profit:float
    retained_earning:float
    has_jobbank_account:bool
    has_bc_employer_certificate:bool

    
class Employee_list(BaseModel):
    employee:str
    job_title:str
    wage:float
    hours_per_week:float
    immigration_status:str	
    employment_start_date:date
    remark:str

class Lmi(BaseModel):
    laid_off_in_12:bool
    laid_off_canadians:Optional[int]
    laid_off_tfw:Optional[int]
    laid_off_reason:Optional[str]
    is_work_sharing:bool
    work_sharing_info:Optional[str]
    labour_dispute:bool
    labour_dispute_info:Optional[str]
    
class LmiaAssess(CommonModel):
    general:General
    employee_list:List[Employee_list]
    lmi:Lmi
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            excels=self.getExcels(["excel/er.xlsx","excel/lmia.xlsx"])
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())

